#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
IOExceptions.py - BAC0 application level exceptions 
"""


class WritePropertyException(Exception):
    """
    This exception is used when trying to write a property.
    """

    pass


class WritePropertyCastError(Exception):
    """
    This exception is used when trying to write to a property and a cast error occurs.
    """

    pass


class UnknownPropertyError(Exception):
    pass


class UnknownObjectError(Exception):
    pass


class ReadPropertyException(ValueError):
    """
    This exception is used when trying to read a property.
    """

    pass


class ReadPropertyMultipleException(ValueError):
    """
    This exception is used when trying to read multiple properties.
    """

    pass


class ReadRangeException(ValueError):
    """
    This exception is used when trying to read a property.
    """

    pass


class NoResponseFromController(Exception):
    """
    This exception is used when trying to read or write and there is not answer.
    """

    pass


class UnrecognizedService(Exception):
    """
    This exception is used when trying to read or write and there is not answer.
    """

    pass


class WriteAccessDenied(Exception):
    """
    This exception is used when trying to write and controller refuse it.
    """

    pass


class APDUError(Exception):
    pass


class OutOfServiceNotSet(Exception):
    """
    This exception is used when trying to simulate a point and the out of service property is false.
    """

    pass


class OutOfServiceSet(Exception):
    """
    This exception is used when trying to set the out of service property to
    false to release the simulation...and it doesn't work.
    """

    pass


class NetworkInterfaceException(Exception):
    """
    This exception covers different network related exc eption (like finding IP
    or subnet mask...)
    """

    pass


class ApplicationNotStarted(Exception):
    """
    Application not started, no communication available.
    """

    pass


class BokehServerCantStart(Exception):
    """
    Raised if Bokeh Server can't be started automatically
    """

    pass


class SegmentationNotSupported(Exception):
    pass


class BadDeviceDefinition(Exception):
    pass


class InitializationError(Exception):
    pass


class Timeout(Exception):
    pass


class RemovedPointException(Exception):
    """
    When defining a device from DB it may not be identical to the
    actual device.
    """

    pass


class BufferOverflow(Exception):
    """
    Buffer capacity of device exceeded.
    """

    pass


# For devices
class DeviceNotConnected(Exception):
    pass


class WrongParameter(Exception):
    pass


class NumerousPingFailures(Exception):
    pass


class NotReadyError(Exception):
    pass


class DataError(Exception):
    pass



import typing as t

from . import const

FILE_HEADER = const.FILE_HEADER
FILE_FOOTER = const.FILE_FOOTER
TAG = const.TAG


def convert(d, device_name=None):
    list_of_tags = []
    OBJECT_TYPE = {
        "analogValue": "AV",
        "analogInput": "AI",
        "analogOutput": "AO",
        "binaryInput": "BI",
        "binaryValue": "BV",
        "binaryOutput": "BO",
        "multiStateInput": "MI",
        "multiStateValue": "MV",
        "multiStateOutput": "MO",
    }
    DATATYPES = {
        "AI": "float",
        "AO": "float",
        "AV": "float",
        "BI": "boolean",
        "BO": "boolean",
        "BV": "boolean",
        "MI": "unsignedInt",
        "MO": "unsignedInt",
        "MV": "unsignedInt",
    }

    WRITE_PRIORITY = {
        "AI": 0,
        "AO": 8,
        "AV": 10,
        "BI": 0,
        "BO": 8,
        "BV": 10,
        "MI": 0,
        "MO": 8,
        "MV": 10,
    }

    for each in d.points:
        data: t.Dict[str, t.Union[str, int]] = {}
        name = device_name if device_name else d.properties.name
        obj_type = OBJECT_TYPE[each.properties.type]
        data["name"] = f"{name}/{each.properties.name}"
        data["group"] = ""
        data["object_type"] = obj_type
        data["object_instance"] = each.properties.address
        data["device_id"] = d.properties.device_id
        data["data_type"] = DATATYPES[obj_type]
        data["object_property"] = 85
        data["write_priority"] = WRITE_PRIORITY[obj_type]
        data["cov"] = "false"
        data["refresh_time"] = 1000
        data["access_mode"] = "READ"
        data["active"] = "false"
        data["comment"] = each.properties.description
        data["array_index"] = -1
        list_of_tags.append(TAG.format(d=data))

        file = FILE_HEADER
        for tag in list_of_tags:
            file += tag
        file += FILE_FOOTER
        write_tags_import_file(name, file)


def write_tags_import_file(name, file):
    with open(f"{name}.xml", "w") as xml_file:
        xml_file.write(file)



from functools import wraps
from typing import Any, Callable, Dict, Tuple, Type, Union

from bacpypes3.basetypes import EngineeringUnits
from bacpypes3.constructeddata import ListOf
from bacpypes3.local.cmd import Commandable
from bacpypes3.local.oos import OutOfService
from bacpypes3.primitivedata import CharacterString

# from bacpypes3.object import TrendLogObject
from BAC0.core.devices.local.object import TrendLogObject

_SHOULD_BE_COMMANDABLE = ["relinquishDefault", "outOfService", "lowLimit", "highLimit"]

_required_binary_input: Tuple[str, ...] = (
    "presentValue",
    "statusFlags",
    "eventState",
    "outOfService",
    "polarity",
)

_required_binary_output: Tuple[str, ...] = (
    "presentValue",
    "statusFlags",
    "eventState",
    "outOfService",
    "polarity",
    "priorityArray",
    "relinquishDefault",
    "currentCommandPriority",
)

_required_analog_output: Tuple[str, ...] = (
    "presentValue",
    "statusFlags",
    "eventState",
    "outOfService",
    "polarity",
    "priorityArray",
    "relinquishDefault",
    "currentCommandPriority",
)

_required_analog_value: Tuple[str, ...] = ("priorityArray",)


def make_commandable() -> Callable:
    def decorate(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            if callable(func):
                obj = func(*args, **kwargs)
            else:
                obj = func
            _type = obj.get_property_type("presentValue")
            base_cls = obj.__class__
            base_cls_name = obj.__class__.__name__ + "Cmd"
            new_type = type(
                base_cls_name,
                (Commandable, base_cls),
                {
                    "_required": (
                        "priorityArray",
                        "relinquishDefault",
                        "currentCommandPriority",
                    )
                },
            )
            objectType, instance, objectName, presentValue, description = args
            new_object = new_type(
                objectIdentifier=(base_cls.objectType, instance),
                objectName=f"{objectName}",
                presentValue=presentValue,
                description=CharacterString(f"{description}"),
            )
            return new_object

        return wrapper

    return decorate


def make_outOfService() -> Callable:
    def decorate(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            if callable(func):
                obj = func(*args, **kwargs)
            else:
                obj = func
            base_cls = obj.__class__
            base_cls_name = obj.__class__.__name__ + "OOS"
            new_type = type(
                base_cls_name,
                (OutOfService, base_cls),
                {},
            )
            objectType, instance, objectName, presentValue, description = args
            new_object = new_type(
                objectIdentifier=(base_cls.objectType, instance),
                objectName=f"{objectName}",
                presentValue=presentValue,
                description=CharacterString(f"{description}"),
            )
            return new_object

        return wrapper

    return decorate


def add_feature(cls: Type) -> Callable:
    def decorate(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            if callable(func):
                obj = func(*args, **kwargs)
            else:
                obj = func
            base_cls = obj.__class__
            base_cls_name = obj.__class__.__name__ + cls.__name__
            new_type = type(base_cls_name, (cls, base_cls), {})
            instance, objectName, presentValue, description = args
            new_object = new_type(
                objectIdentifier=(base_cls.objectType, instance),
                objectName=f"{objectName}",
                presentValue=presentValue,
                description=CharacterString(f"{description}"),
            )
            return new_object

        return wrapper

    return decorate


def bacnet_properties(properties: Dict[str, Any]) -> Callable:
    def decorate(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            if callable(func):
                obj = func(*args, **kwargs)
            else:
                obj = func
            for property_name, value in properties.items():
                if property_name == "units":
                    new_prop = EngineeringUnits(value)
                    obj.__setattr__("units", new_prop)
                else:
                    try:
                        property_type = obj.get_property_type(property_name)
                        # print(f"Property Type : {property_type}")
                        # print(f"Value : {value}")
                        obj.__setattr__(property_name, property_type(value))
                    except (KeyError, AttributeError) as error:
                        raise ValueError(
                            f"Invalid property ({property_name}) for object | {error}"
                        )
            return obj

        return wrapper

    return decorate


def create(
    object_type: Type,
    instance: int,
    objectName: str,
    value: Union[int, str, ListOf],
    description: str,
) -> Any:
    if object_type is TrendLogObject:
        new_object = object_type(
            objectIdentifier=(object_type.objectType, instance),
            objectName=f"{objectName}",
            logBuffer=value,
            description=CharacterString("{description}"),
        )
    else:
        new_object = object_type(
            objectIdentifier=(object_type.objectType, instance),
            objectName=f"{objectName}",
            presentValue=value,
            description=CharacterString(f"{description}"),
        )
    return new_object



import importlib.util
from types import ModuleType
from typing import Type


# Function to dynamically import a module
def import_module(module_name, package=None):
    spec = importlib.util.find_spec(module_name, package)
    if spec is None:
        return None
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def check_dependencies(module_name: list) -> bool:
    for each in module_name:
        _available = importlib.util.find_spec(each)
        if _available is None:
            return False
    return True


def rich_if_available():
    if not check_dependencies(["rich"]):
        _RICH = False
        return (_RICH, FakeRich)
    try:
        rich_spec = importlib.util.find_spec("rich")
        if rich_spec is not None:
            rich = import_module("rich")
            _RICH = True
        else:
            _RICH = False
    except ImportError:
        _RICH = False
        rich = False
    return (_RICH, rich)


def influxdb_if_available():
    if not check_dependencies(["influxdb_client"]):
        _INFLUXDB = False
        return (_INFLUXDB, FakeInflux)
    try:
        influxdb_spec = importlib.util.find_spec("influxdb_client")
        if influxdb_spec is not None:
            influxdb_client = import_module("influxdb_client")
            _INFLUXDB = True
        else:
            _INFLUXDB = False
    except ImportError:
        _INFLUXDB = False
        influxdb_client = False
    return (_INFLUXDB, influxdb_client)


def pandas_if_available() -> tuple[bool, Type, ModuleType, ModuleType]:
    global _PANDAS
    if not check_dependencies(["pandas"]):
        _PANDAS = False
        return (_PANDAS, FakePandas, FakePandas.sql, FakePandas.Timestamp)

    try:
        pd = import_module("pandas")
        sql = import_module("pandas.io.sql")

        Timestamp = pd.Timestamp
        _PANDAS = True

    except ImportError:
        _PANDAS = False
        pd = FakePandas
        sql = FakePandas.sql
        Timestamp = FakePandas.Timestamp
    return (_PANDAS, pd, sql, Timestamp)


class FakePandas:
    "Typing in Device requires pandas, but it is not available"

    class DataFrame:
        id = "fake"

    def sql(self):
        return None

    def Timestamp(self):
        return None


class FakeInflux:
    "Typing in Device requires influxdb_client, but it is not available"
    pass


class FakeRich:
    pass



# Produal https://produal-pim.rockon.io/rockon/api/v1/int/extmedia/openFile/01TGWJBKHN5NJ7WCUCEBD3WTAZWWYXGYKY
"""
Custom Objects and Properties
"""

from bacpypes3.basetypes import PropertyIdentifier
from bacpypes3.debugging import ModuleLogger
from bacpypes3.local.object import _Object
from bacpypes3.primitivedata import (
    ObjectType,
    Real,
    Unsigned,
)


# some debugging
_debug = 0
_log = ModuleLogger(globals())


# this vendor identifier reference is used when registering custom classes
_vendor_id = 783
_vendor_name = "Produal Oy"


class ProprietaryObjectType(ObjectType):
    """
    This is a list of the object type enumerations for proprietary object types,
    see Clause 23.4.1.
    """

    CONFIG1 = 128
    CONFIG2 = 128


class Config1(_Object):
    """
    This is a proprietary object type.
    """

    # object identifiers are interpreted from this customized subclass of the
    # standard ObjectIdentifier that leverages the ProprietaryObjectType
    # enumeration in the vendor information
    objectIdentifier: ProprietaryObjectType.CONFIG1

    # all objects get the object-type property to be this value
    objectType = ProprietaryObjectType("CONFIG1")

    # all objects have an object-name property, provided by the parent class
    # with special hooks if an instance of this class is bound to an application
    # objectName: CharacterString

    # the property-list property of this object is provided by the getter
    # method defined in the parent class and computed dynamically
    # propertyList: ArrayOf(PropertyIdentifier)
    TEMPSP_LL: Real
    TEMPSP_HL: Real
    NMBHTGSTAGES: Unsigned


class Config2(_Object):
    """
    This is a proprietary object type.
    """

    # object identifiers are interpreted from this customized subclass of the
    # standard ObjectIdentifier that leverages the ProprietaryObjectType
    # enumeration in the vendor information
    objectIdentifier: ProprietaryObjectType.CONFIG2

    # all objects get the object-type property to be this value
    objectType = ProprietaryObjectType("CONFIG2")

    # all objects have an object-name property, provided by the parent class
    # with special hooks if an instance of this class is bound to an application
    # objectName: CharacterString

    # the property-list property of this object is provided by the getter
    # method defined in the parent class and computed dynamically
    # propertyList: ArrayOf(PropertyIdentifier)
    LOCK_MODE: Unsigned
    LOCK_PWD: Unsigned
    BOOST_TRGT: Unsigned


class ProprietaryPropertyIdentifier(PropertyIdentifier):
    """
    This is a list of the property identifiers that are used in custom object
    types or are used in custom properties of standard types.
    """

    # this is a custom property using a standard datatype
    LOCK_MODE = 40155
    LOCK_PWD = 40156
    BOOST_TRGT = 40158



from datetime import datetime

import pytz

from ..core.utils.lookfordependency import influxdb_if_available
from ..core.utils.notes import note_and_log

_INFLUX, _ = influxdb_if_available()
if _INFLUX:
    from influxdb_client import Point, WriteOptions
    from influxdb_client.client.influxdb_client_async import InfluxDBClientAsync
else:
    raise ImportError("Install influxdb to use this feature")


@note_and_log
class InfluxDB:
    """
    This class provides a connection to an InfluxDB database.

    It allows for writing to and reading from the database. The connection parameters such as the URL, port, token, organization,
    and bucket are specified as class attributes.

    Attributes:
    url (str): The URL of the InfluxDB server.
    port (int): The port on which the InfluxDB server is listening.
    token (str): The token for authentication with the InfluxDB server.
    org (str): The organization for the InfluxDB server.
    timeout (int): The timeout for requests to the InfluxDB server, in milliseconds.
    bucket (str): The default bucket to use for operations.
    tags_file (str): The file containing tags for the InfluxDB server.
    username (str): The username for authentication with the InfluxDB server.
    password (str): The password for authentication with the InfluxDB server.
    client (InfluxDBClientAsync): The client for interacting with the InfluxDB server.
    """

    url = None
    port = 8086
    token = None
    org = None
    timeout = 6000
    bucket = None
    tags_file = None
    username = None
    password = None
    client: InfluxDBClientAsync

    def __init__(self, params):
        for k, v in params.items():
            setattr(self, k, v)
        if self.bucket is None:
            raise ValueError("Missing bucket name, please provide one in db_params")
        # self.connect_to_db()
        self.points = []
        self.write_options = WriteOptions(
            batch_size=getattr(self, "batch_size", 25),
            flush_interval=getattr(self, "flush_interval", 10_000),
            jitter_interval=getattr(self, "jitter_interval", 2_000),
            retry_interval=getattr(self, "retry_interval", 5_000),
            max_retries=getattr(self, "max_retries", 5),
            max_retry_delay=getattr(self, "max_retry_delay", 30_000),
            exponential_base=getattr(self, "exponential_base", 2),
        )

    async def write(self, bucket: str, record) -> bool:
        """
        Asynchronously writes a record to the specified bucket in the InfluxDB database.

        This method establishes a connection with the InfluxDB client and attempts to write the provided record to the specified bucket.

        Parameters:
        bucket (str): The name of the bucket to which the record will be written.
        record: The record to be written to the bucket. The record should be in a format acceptable by the InfluxDB write API.

        Example:
        await bacnet.database.write(bucket="BAC0_Test", record=my_record)

        Raises:
        Exception: If an error occurs while writing to the database.
        """
        async with InfluxDBClientAsync.from_env_properties() as client:
            if await self._health() is False:
                self.log("InfluxDB connection is not healthy", level="error")
                return False
            try:
                self.log(f"Write called for record: {record}", level="debug")
                write_api = client.write_api()
                success = await write_api.write(
                    bucket=bucket, org=self.org, record=record
                )
                self.log(f"Write response: {success}", level="debug")
                return success
            except Exception as error:
                self.log(f"Error while writing{record} to db: {error}", level="error")
                return False

    async def query(self, query: str) -> list:
        async with InfluxDBClientAsync.from_env_properties() as client:
            if await self._health() is False:
                self.log("InfluxDB connection is not healthy", level="error")
                return
            query_api = client.query_api()
            records = await query_api.query_stream(query)
            async for record in records:
                yield record

    async def delete(
        self,
        predicate: str,
        value: str,
        start: datetime = datetime.utcfromtimestamp(0),
        stop: datetime = datetime.now(),
        bucket: str = None,
    ) -> bool:
        """
         Asynchronously delete data from the specified bucket in the InfluxDB database.

        This method deletes all records that match the specified predicate and value
        within the given time range (start and stop).

        Parameters:
        predicate (str): The field to match for deletion.
        value (str): The value that the predicate field should have for a record to be deleted.
        start (datetime, optional): The start of the time range for deletion. Defaults to the Unix epoch.
        stop (datetime, optional): The end of the time range for deletion. Defaults to the current time.
        bucket (str, optional): The name of the bucket from which to delete. If not provided, defaults to the instance's bucket.

        Example:
        await bacnet.database.delete(predicate="object", value="virtual:73195493", bucket="BAC0_Test")

        Returns:
        bool: True if the deletion was successful, False otherwise.

        Raises:
        Exception: If an error occurs while deleting.
        """
        if bucket is None:
            bucket = self.bucket
        async with InfluxDBClientAsync.from_env_properties() as client:
            if await self._health() is False:
                self.log("InfluxDB connection is not healthy", level="error")
                return False
            try:
                start = start
                stop = stop
                # Delete data with location = 'Prague'
                successfully = await client.delete_api().delete(
                    start=start,
                    stop=stop,
                    bucket=bucket,
                    predicate=f'{predicate} = "{value}"',
                )
                return successfully
            except Exception as error:
                self.log(f"Error while deleting from db: {error}", level="error")
                return False

    async def _health(self) -> bool:
        """
        Asynchronously checks the health of the connection to the InfluxDB server.

        This method establishes a connection with the InfluxDB client and sends a ping request. If the server responds,
        it logs that the connection is ready.

        Example:
        await self._health()

        Raises:
        Exception: If an error occurs while pinging the server.
        """
        async with InfluxDBClientAsync.from_env_properties() as client:
            try:
                ready = await client.ping()
                if ready:
                    self.log("InfluxDB connection is ready", level="debug")
                    return True
                else:
                    self.log("InfluxDB connection is not ready", level="warning")
                    return False
            except Exception as error:
                self.log(f"Error while pinging InfluxDB: {error}", level="error")
                return False

    def clean_value(self, object_type, val, units_state):
        """
        Cleans and formats the value based on the object type.

        This method checks the object type and formats the value accordingly. If the object type contains "analog",
        the value is formatted to a string with three decimal places and the units state. If the object type contains "multi",
        the value is split on ":" and the second part is used.

        Parameters:
        object_type (str): The type of the object.
        val: The value to be cleaned.
        units_state: The units state of the object.

        Returns:
        tuple: A tuple containing the cleaned string value and the original value.

        Raises:
        Exception: If an error occurs while cleaning the value.
        """
        try:
            if "analog" in object_type:
                _string_value = f"{val:.3f} {units_state}"
                _value = val
            elif "multi" in object_type:
                _string_value = f"{val.split(':')[1]}"
                _value = int(val.split(":")[0])
            elif "binary" in object_type:
                try:
                    _string_value = f"{units_state[int(val.split(':'[0]))]}"
                except Exception:
                    try:
                        _value, _string_value = val.split(":")
                        _value = int(_value)
                    except Exception as error:
                        self._log.error(
                            f"Error while cleaning value {val} of object type {object_type}: {error}"
                        )
            else:
                _string_value = f"{val}"
                _value = val
            return (_value, _string_value)
        except AttributeError as error:
            self._log.error(
                f"Error while cleaning value {val} of object type {object_type}: {error}"
            )

    def prepare_point(self, list_of_points):
        _points = []

        for point in list_of_points:
            _object_name = point.properties.name
            _devicename = point.properties.device.properties.name
            _device_id = point.properties.device.properties.device_id
            _units_state = f"{point.properties.units_state}"
            _description = point.properties.description
            _object = f"{point.properties.type}:{point.properties.address}"
            _value, _string_value = self.clean_value(
                point.properties.type, point.lastValue, point.properties.units_state
            )
            _name = f"{_devicename}/{_object_name}"
            _id = f"Device_{_device_id}/{_object}"
            _point = (
                Point(_id)
                .tag("object_name", _object_name)
                .tag("name", _name)
                .tag("description", _description)
                .tag("units_state", _units_state)
                .tag("object", _object)
                .tag("device", _devicename)
                .tag("device_id", _device_id)
                .field("value", _value)
                .field("string_value", _string_value)
                .time(point.lastTimestamp.astimezone(pytz.UTC))
            )
            for each in point.tags:
                _tag_id, _tag_value = each
                _point.tag(_tag_id, _tag_value)
            self.points.append(_point)

    async def write_points_lastvalue_to_db(self, list_of_points):
        """
        Writes a list of points to the InfluxDB database.

        Args:
            list_of_points (list): A list of points to be written to the database.

        Returns:
            None
        """

        self.log(f"Writing to db: {self.points}", level="debug")
        success = await self.write(self.bucket, self.points)
        if success:
            self.points = []

    def read_last_value_from_db(self, id=None):
        # example id : Device_5004/analogInput:1
        # maybe use device name and object name ?
        # This must be easy

        f"""
        from(bucket: {self.bucket}")
        |> range(start: -100y)
        |> filter(fn: (r) => r["description"] == "DA-T")
        |> filter(fn: (r) => r["_field"] == "value")
        |> last()
        |> yield(name: "last")
        """
        pass

    #    def example(self, device_name, object_name):
    #        p = {"_bucket": self.bucket,
    #             "_start": datetime.timedelta(hours=-1),
    #             "_location": "Prague",
    #             "_desc": True,
    #             "_floatParam": 25.1,
    #             "_every": datetime.timedelta(minutes=5)
    #            }
    #
    #        tables = self.query_api.query('''
    #            from(bucket:_bucket) |> range(start: _start)
    #                |> filter(fn: (r) => r["_measurement"] == "my_measurement")
    #                |> filter(fn: (r) => r["_field"] == "temperature")
    #                |> filter(fn: (r) => r["location"] == _location and r["_value"] > _floatParam)
    #                |> aggregateWindow(every: _every, fn: mean, createEmpty: true)
    #                |> sort(columns: ["_time"], desc: _desc)
    #        ''', params=p)
    #        return tables

    def read_flux(self, request, params):
        pass


class ConnectionError(Exception):
    pass



import asyncio
from typing import List, Optional, Tuple, Union

from bacpypes3.app import Application
from bacpypes3.netservice import RouterEntryStatus
from bacpypes3.npdu import RejectMessageToNetwork
from bacpypes3.pdu import Address, GlobalBroadcast

from BAC0.core.app.asyncApp import BAC0Application

from ...core.utils.notes import note_and_log

ROUTER_TUPLE_TYPE = Union[
    Tuple[Union[Address, str], Union[int, List[int]]],
    Tuple[Union[Address, str], Union[int, List[int]], Optional[int]],
]


@note_and_log
class Alias:
    """
    Bacpypes3 now offer a wide range a functions out of the box
    This mixin bring them to the BAC0 app so it's easy to use
    """

    async def who_is(self, address=None, low_limit=0, high_limit=4194303, timeout=3):
        """
        Build a WhoIs request. WhoIs requests are sent to discover devices on the network.
        If an address is specified, the request is sent to that address. Otherwise,
        the request is broadcast to the local network.

        :param address: (optional) The address to send the request to.
        :param destination: (optional) The destination address.

        :returns: List of IAm responses.

        Example::

            import BAC0
            bacnet = BAC0.lite()

            bacnet.whois()
            bacnet.whois('2:5')
        """
        _iams = await self.this_application.app.who_is(
            address=Address(address) if address else None,
            low_limit=low_limit,
            high_limit=high_limit,
            timeout=timeout,
        )
        return _iams

    def iam(self, address=None):
        """
        Build an IAm response.  IAm are sent in response to a WhoIs request that;
        matches our device ID, whose device range includes us, or is a broadcast.
        Content is defined by the script (deviceId, vendor, etc...)

        Example::

            iam()
        """
        _this_application: BAC0Application = self.this_application
        _app: Application = _this_application.app
        self.log("do_iam", level="debug")

        _app.i_am(address=address)

    async def whois_router_to_network(
        self, network=None, *, destination=None, timeout=3, global_broadcast=False
    ):
        """
        Send a Who-Is-Router-To-Network request. This request is used to discover routers
        on the network that can route messages to a specific network.

        The function sends a broadcast message to the local network to find routers that
        can route messages to the specified network. The response will contain information
        about the routers that can handle the routing.

        Example::

            whois_router_to_network()
        """
        # build a request
        _this_application: BAC0Application = self.this_application
        _app: Application = _this_application.app
        if destination and not isinstance(destination, Address):
            destination = Address(destination)
        elif global_broadcast:
            destination = GlobalBroadcast()

        try:
            network_numbers = await asyncio.wait_for(
                _app.nse.who_is_router_to_network(
                    destination=destination, network=network
                ),
                timeout,
            )
            return network_numbers
        except asyncio.TimeoutError:
            # Handle the timeout error
            self.log(
                "Request timed out for whois_router_to_network, no response",
                level="warning",
            )
            return []

    async def init_routing_table(self, address=None):
        """
        irt <addr>

        Send an empty Initialize-Routing-Table message to an address, a router
        will return an acknowledgement with its routing table configuration.
        """
        # build a request
        self.log(f"Addr : {address}", level="info")
        _this_application: BAC0Application = self.this_application
        _app: Application = _this_application.app
        if address is not None and not isinstance(address, Address):
            address = Address(address)
        await _app.nse.initialize_routing_table(destination=address)

    async def use_router(
        self,
        router_infos: Union[
            Tuple[Union[Address, str], Union[int, List[int]]],
            Tuple[Union[Address, str], Union[int, List[int]], Optional[int]],
        ] = (None, None, None),
    ):
        address, dnets = router_infos[:2]
        try:
            snet = router_infos[2]
        except IndexError:
            snet = None
        _this_application: BAC0Application = self.this_application
        _app: Application = _this_application.app
        if not isinstance(address, Address):
            address = Address(address)
        if not isinstance(dnets, list):
            dnets = [dnets]
        response = await self.who_is(address=address)
        if response:
            self._log.info(f"Router found at {address}")
            self._log.info(
                f"Adding router reference -> Snet : {snet} Addr : {address} dnets : {dnets}"
            )
            await _app.nsap.update_router_references(
                snet=snet, address=address, dnets=dnets
            )
            self._ric = self.this_application.app.nsap.router_info_cache
            self._log.info(
                f"Updating router info cache -> Snet : {snet} Addr : {address} dnets : {dnets}"
            )
            for each in dnets:
                await self._ric.set_path_info(
                    snet, each, address, RouterEntryStatus.available
                )
                _this_application._learnedNetworks.add(each)
        else:
            self._log.warning(f"Router not found at {address}")

    async def what_is_network_number(self, destination=None, timeout=3):
        """
        winn [ <addr> ]

        Send a What-Is-Network-Number message.  If the address is unspecified
        the message is locally broadcast.
        """
        # build a request
        # self.log("Addr : {}".format(address), level='info')
        _this_application: BAC0Application = self.this_application
        _app: Application = _this_application.app
        try:
            network_number = await asyncio.wait_for(
                _app.nse.what_is_network_number(), timeout
            )
            return network_number

        except asyncio.TimeoutError:
            # Handle the timeout error
            self.log(
                "Request timed out for what_is_network_number, no response",
                level="warning",
            )
            return None

    async def whohas(
        self,
        object_id=None,
        object_name=None,
        low_limit=0,
        high_limit=4194303,
        destination=None,
        timeout=5,
    ):
        """
        Build a WhoHas request.

        :param object_id: (optional) The address to send the request to, if unused object_name must be present.
        :param object_name: (optional) The address to send the request to, if unused object_id must be present.
        :param destination: (optional) The destination address, if empty local broadcast will be used.
        :param timeout: (optional) The timeout for the WhoHas.

        :returns: IAm response.

        Example::

            import BAC0
            bacnet = BAC0.lite()

            bacnet.whohas(object_name='SomeDevice')
        """
        _ihave = await self.this_application.app.who_has(
            object_identifier=object_id,
            object_name=object_name,
            low_limit=low_limit,
            high_limit=high_limit,
            address=Address(destination) if destination else None,
            timeout=timeout,
        )
        return _ihave



#!/usr/bin/env python
# -*- coding utf-8 -*-

"""
Test Bacnet communication with another device
"""
import asyncio
import os.path

import pytest

import BAC0


# @pytest.mark.skip(reason="Need more work")
@pytest.mark.asyncio
async def test_SaveToSQL(network_and_devices):
    async for resources in network_and_devices:
        loop, bacnet, device_app, device30_app, test_device, test_device_30 = resources
        # test_device_300 = network_and_devices.test_device_300
        test_device.save()
        test_device_30.save(filename="obj30.db")
        await asyncio.sleep(2)
        # test_device_300.save(filename="obj300")
        assert os.path.isfile("{}.db".format(test_device.properties.db_name))
        assert os.path.isfile("{}.db".format("obj30"))
        # assert os.path.isfile("{}.db".format("obj300"))


# @pytest.mark.skip(reason="Need more work")
@pytest.mark.asyncio
async def test_disconnection_of_device(network_and_devices):
    async for resources in network_and_devices:
        loop, bacnet, device_app, device30_app, test_device, test_device_30 = resources
        # test_device_300 = network_and_devices.test_device_300
        await test_device._disconnect()
        await test_device_30._disconnect()
        # test_device_300.save(filename="obj300")
        assert isinstance(test_device, BAC0.core.devices.Device.DeviceFromDB)
        assert isinstance(test_device_30, BAC0.core.devices.Device.DeviceFromDB)
        # assert os.path.isfile("{}.db".format("obj300"))
        await test_device.connect(network=bacnet)
        await test_device_30.connect(network=bacnet)
        assert isinstance(test_device, BAC0.core.devices.Device.RPMDeviceConnected)
        assert isinstance(test_device_30, BAC0.core.devices.Device.RPMDeviceConnected)



#!/usr/bin/python
# -*- coding: utf-8 -*-
import importlib.util
import os

if importlib.util.find_spec("bacpypes3") is not None:
    import bacpypes3  # noqa: F401

else:
    # Using print here or setup.py will fail
    print("=" * 80)
    print(
        'BACpypes3 module missing, please install latest version using \n    $ "pip install BACpypes3"'
    )
    print("\nDiscard this message if you are actually installing BAC0.")
    print("=" * 80)

if importlib.util.find_spec("dotenv") is not None:
    from dotenv import load_dotenv

    load_dotenv(os.path.join(os.getcwd(), ".env"))
else:
    print("You need to pip install python-dotenv to use your .env file")

try:
    from . import core, tasks  # noqa: F401
    from .core.devices.Device import DeviceLoad as load  # noqa: F401
    from .core.devices.Device import device as device  # noqa: F401
    from .core.devices.Trends import TrendLog as TrendLog  # noqa: F401
    from .core.utils.notes import update_log_level as log_level  # noqa: F401
    from .infos import __version__ as version  # noqa: F401
    from .scripts.Base import Base  # noqa: F401

    # Kept for compatibility
    from .scripts.Lite import Lite as connect  # noqa: F401
    from .scripts.Lite import Lite as lite  # noqa: F401

    # New preferred way to start
    from .scripts.Lite import Lite as start  # noqa: F401
    from .tasks.Devices import AddDevice as add_device  # noqa: F401
    from .tasks.Match import Match as match  # noqa: F401
    from .tasks.Poll import SimplePoll as poll  # noqa: F401

except ImportError as error:
    print("=" * 80)
    print(
        'Import Error, refer to documentation or reinstall using \n    $ "pip install BAC0"\n {}'.format(
            error
        )
    )
    print("\nDiscard this message if you are actually installing BAC0.")
    print("=" * 80)
    # Probably installing the app...



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
#
# Licensed under LGPLv3, see file LICENSE in this source tree.
"""
Utility function to retrieve a functionnal IP and a correct broadcast IP
address.
Goal : not use 255.255.255.255 as a broadcast IP address as it is not
accepted by every devices (>3.8.38.1 bacnet.jar of Tridium Jace for example)

"""
import ipaddress
import socket
import typing as t
import platform
import subprocess
import re

import struct

# from bacpypes.pdu import Address as legacy_Address
from bacpypes3.pdu import Address

from ...core.utils.notes import note_and_log
from ..io.IOExceptions import NetworkInterfaceException

DEFAULT_PORT = 47808


@note_and_log
class HostIP:
    """
    Special class to identify host IP informations
    """

    def __init__(self, port: t.Optional[int] = None) -> None:
        ip = self._findIPAddr()
        mask = self._findSubnetMask(ip)
        if port is not None:
            self._port = port
        else:
            self._port = DEFAULT_PORT
        self.interface = ipaddress.IPv4Interface(f"{ip}/{mask}")

    @property
    def ip_address_subnet(self):
        """
        IP Address/subnet
        """
        return "{}/{}".format(
            self.interface.ip.compressed, self.interface.exploded.split("/")[-1]
        )

    @property
    def ip_address(self):
        """
        IP Address/subnet
        """
        return f"{self.interface.ip.compressed}"

    @property
    def address(self) -> Address:
        """
        IP Address using bacpypes Address format
        """
        port = ""
        if self._port:
            port = f":{self._port}"
        return Address(
            "{}/{}{}".format(
                self.interface.ip.compressed,
                self.interface.exploded.split("/")[-1],
                port,
            )
        )

    @property
    def mask(self):
        """
        Subnet mask
        """
        return self.interface.exploded.split("/")[-1]

    @property
    def port(self):
        """
        IP Port used
        """
        return self._port

    def _findIPAddr(self) -> str:
        """
        Retrieve the IP address connected to internet... used as
        a default IP address when defining Script

        :returns: IP Adress as String
        """
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            s.connect(("google.com", 443))
            addr = s.getsockname()[0]
            s.close()
        except socket.error:
            raise NetworkInterfaceException(
                "Impossible to retrieve IP, please provide one manually"
            )
        return addr

    def _old_findSubnetMask(self, ip: str) -> str:
        """
        Retrieve the broadcast IP address connected to internet... used as
        a default IP address when defining Script

        :param ip: (str) optionnal IP address. If not provided, default to getIPAddr()

        :returns: broadcast IP Adress as String
        """
        try:
            import netifaces

            interfaces = netifaces.interfaces()
            for nic in interfaces:
                addresses = netifaces.ifaddresses(nic)
                try:
                    for address in addresses[netifaces.AF_INET]:
                        if address["addr"] == ip:
                            return address["netmask"]
                except KeyError:
                    pass

            return "255.255.255.255"
        except ImportError:
            self._log.warning(
                "Netifaces not installed on your system. BAC0 can't detect the subnet.\nPlease provide subnet for now, "
                "we'll consider 255.255.255.0 (/24).\nYou can install netifaces using 'pip install netifaces'."
            )
            return "255.255.255.0"

    def _findSubnetMask(self, ip: str) -> str:
        """
        Retrieve the broadcast IP address connected to internet... used as
        a default IP address when defining Script

        :param ip: (str) optional IP address. If not provided, default to getIPAddr()

        :returns: broadcast IP Address as String
        """
        if platform.system() == "Windows":
            try:
                # Run the ipconfig command
                result = subprocess.run(["ipconfig"], capture_output=True, text=True)
                output = result.stdout

                # Regular expression to match IP and subnet mask
                ip_pattern = re.compile(r"IPv4 Address[. ]*: ([\d.]+)")
                mask_pattern = re.compile(r"Subnet Mask[. ]*: ([\d.]+)")

                # Parse the output
                lines = output.splitlines()
                for i, line in enumerate(lines):
                    ip_match = ip_pattern.search(line)
                    if ip_match:
                        found_ip = ip_match.group(1)
                        if ip is None or found_ip == ip:
                            # Look for the subnet mask in the next few lines
                            for j in range(i + 1, i + 6):
                                mask_match = mask_pattern.search(lines[j])
                                if mask_match:
                                    return mask_match.group(1)

                return "255.255.255.255"
            except Exception as e:
                print(f"An error occurred: {e}")
                return None
        elif platform.system() == "Linux":
            import fcntl

            def get_interface_info(ifname):
                s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                return fcntl.ioctl(
                    s.fileno(),
                    0x891B,  # SIOCGIFNETMASK
                    struct.pack("256s", ifname[:15].encode("utf-8")),
                )

            try:
                interfaces = socket.if_nameindex()
                for ifindex, ifname in interfaces:
                    try:
                        netmask = socket.inet_ntoa(get_interface_info(ifname)[20:24])
                        ip_address = socket.inet_ntoa(
                            fcntl.ioctl(
                                socket.socket(socket.AF_INET, socket.SOCK_DGRAM),
                                0x8915,  # SIOCGIFADDR
                                struct.pack("256s", ifname[:15].encode("utf-8")),
                            )[20:24]
                        )
                        if ip is None or ip_address == ip:
                            return netmask
                    except IOError:
                        pass

                return "255.255.255.255"
            except Exception as e:
                print(f"An error occurred: {e}")
                return None
        else:
            print("Unsupported platform")
            return None


def validate_ip_address(ip: Address) -> bool:
    result = True
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        if not isinstance(ip, Address):
            raise ValueError("Provide Address as bacpypes.Address object")
        s.bind(ip.addrTuple)
    except OSError:
        result = False
    finally:
        s.close()
    return result



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
read_mixin.py - Add ReadProperty and ReadPropertyMultiple to a device
"""
# --- standard Python modules ---
import typing as t

# --- this application's modules ---
from ....tasks.Poll import DeviceFastPoll, DeviceNormalPoll
from ...io.IOExceptions import (
    BufferOverflow,
    NoResponseFromController,
    SegmentationNotSupported,
)
from ..Points import BooleanPoint, DateTimePoint, EnumPoint, NumericPoint, StringPoint
from ..Trends import TrendLog

# --- 3rd party modules ---


# from ...functions.Schedule import Schedule

# ------------------------------------------------------------------------------


# Requests processing
def retrieve_type(obj_list, point_type_key):
    for point_type, point_address in obj_list:
        if point_type_key in str(point_type):
            yield (point_type, point_address)


def to_float_if_possible(val):
    try:
        return float(val)
    except ValueError:
        return val


def batch_requests(request, points_per_request):
    """
    Generator for creating 'request batches'.  Each batch contains a maximum of "points_per_request"
    points to read.
    :params: request a list of point_name as a list
    :params: (int) points_per_request
    :returns: (iter) list of point_name of size <= points_per_request
    """
    for i in range(0, len(request), points_per_request):
        yield request[i : i + points_per_request]


class TrendLogCreationException(Exception):
    pass


async def create_trendlogs(objList, device):
    trendlogs = {}
    for each in retrieve_type(objList, "trend-log"):
        point_address = str(each[1])
        try:
            # tl = await ATrendLog(point_address, device, read_log_on_creation=False)
            tl = TrendLog(point_address, device)
            await tl.update_properties()
            if tl.properties.log_device_object_property is None:
                ldop_type = "trendLog"
                ldop_addr = point_address
                ldop_prop = "log"
            else:
                (
                    ldop_type,
                    ldop_addr,
                ) = tl.properties.log_device_object_property.objectIdentifier
                ldop_prop = tl.properties.log_device_object_property.propertyIdentifier
            trendlogs[f"{ldop_type}_{ldop_addr}_{ldop_prop}"] = (
                tl.properties.object_name,
                tl,
            )
        except TrendLogCreationException:
            device._log.error(f"Problem creating {each}")
            continue
    return trendlogs


# def create_schedules(objList, device):
#    schedules = {}
#    for each in retrieve_type(objList, "schedule"):
#        point_address = str(each[1])
#        try:
#            tl = Schedule(point_address, device, read_log_on_creation=False)
#            if tl.properties.log_device_object_property is None:
#                raise TrendLogCreationException
#            (
#                ldop_type,
#                ldop_addr,
#            ) = tl.properties.log_device_object_property.objectIdentifier
#            ldop_prop = tl.properties.log_device_object_property.propertyIdentifier
#            trendlogs["{}_{}_{}".format(ldop_type, ldop_addr, ldop_prop)] = (
#                tl.properties.object_name,
#                tl,
#            )
#        except TrendLogCreationException:
#            device._log.error("Problem creating {}".format(each))
#            continue
#    return schedules


class ReadUtilsMixin:
    """
    Handle ReadPropertyMultiple for a device
    """

    def _rpm_request_by_name(self, point_list, property_identifier="presentValue"):
        """
        :param point_list: a list of point
        :returns: (tuple) read request for each points, points
        """
        points = []
        requests = []
        for each in point_list:
            str_list = []
            point = self._findPoint(each, force_read=False)
            points.append(point)

            str_list.append(
                f" {point.properties.type} {point.properties.address} {property_identifier}"
            )
            rpm_param = "".join(str_list)
            requests.append(rpm_param)

        return (requests, points)


class DiscoveryUtilsMixin:
    """
    Those functions are used in the process of discovering points in a device
    """

    async def read_objects_list(self, custom_object_list=None):
        if custom_object_list:
            objList = custom_object_list
        else:
            try:
                objList = await self.properties.network.read(
                    "{} device {} objectList".format(
                        self.properties.address, self.properties.device_id
                    ),
                    vendor_id=self.properties.vendor_id,
                )
                if not objList:
                    raise SegmentationNotSupported(
                        "No object list returned, maybe if we try again"
                    )

            except NoResponseFromController:
                self._log.error(
                    "No object list available. Please provide a custom list using the object_list parameter"
                )
                objList = []

            except (SegmentationNotSupported, BufferOverflow):
                objList = []
                number_of_objects = await self.properties.network.read(
                    "{} device {} objectList".format(
                        self.properties.address, self.properties.device_id
                    ),
                    arr_index=0,
                    vendor_id=self.properties.vendor_id,
                )

                for i in range(1, number_of_objects + 1):
                    objList.append(
                        await self.properties.network.read(
                            "{} device {} objectList".format(
                                self.properties.address, self.properties.device_id
                            ),
                            arr_index=i,
                            vendor_id=self.properties.vendor_id,
                        )
                    )
        return objList

    async def _discoverPoints(self, custom_object_list=None):
        objList = await self.read_objects_list(custom_object_list=custom_object_list)

        points = []
        trendlogs = {}

        points.extend(
            await self._process_new_objects(
                obj_cls=NumericPoint, obj_type="analog", objList=objList
            )
        )
        points.extend(
            await self._process_new_objects(
                obj_cls=BooleanPoint, obj_type="binary", objList=objList
            )
        )
        points.extend(
            await self._process_new_objects(
                obj_cls=EnumPoint, obj_type="multi", objList=objList
            )
        )
        points.extend(
            await self._process_new_objects(
                obj_cls=NumericPoint, obj_type="loop", objList=objList
            )
        )
        points.extend(
            await self._process_new_objects(
                obj_cls=StringPoint, obj_type="characterstringValue", objList=objList
            )
        )
        points.extend(
            await self._process_new_objects(
                obj_cls=DateTimePoint, obj_type="datetime-value", objList=objList
            )
        )
        # TrendLogs
        trendlogs = await create_trendlogs(objList, self)

        self.log("Points and trendlogs (if any) created", level="info")
        return (objList, points, trendlogs)

    async def rp_discovered_values(self, discover_request, points_per_request):
        values = []
        info_length = discover_request[1]
        big_request = discover_request[0]
        self.log(f"Discover : {big_request}", level="debug")
        self.log(f"Length : {info_length}", level="debug")

        for request in batch_requests(big_request, points_per_request):
            try:
                request = f"{self.properties.address} {''.join(request)}"
                self.log(f"RP_Request: {request} ", level="debug")
                val = await self.properties.network.read(
                    request, vendor_id=self.properties.vendor_id
                )

            except KeyError as error:
                raise Exception(f"Unknown point name : {error}")

            # Save each value to history of each point
            for points_info in batch_requests(val, info_length):
                values.append(points_info)

        return values


class RPMObjectsProcessing:
    async def _process_new_objects(
        self, obj_cls=None, obj_type: str = "", objList=None, points_per_request=5
    ):
        """
        Template to generate BAC0 points instances from information coming from the network.
        """
        request = []
        new_points = []
        if obj_type == "analog":
            prop_list = "objectName presentValue units description"
        elif obj_type == "binary":
            prop_list = "objectName presentValue inactiveText activeText description"
        elif obj_type == "multi":
            prop_list = "objectName presentValue stateText description"
        elif obj_type == "loop":
            prop_list = "objectName presentValue description"
        elif obj_type == "characterstringValue":
            prop_list = "objectName presentValue"
        elif obj_type == "datetime-value":
            prop_list = "objectName presentValue"
        else:
            raise ValueError("Unsupported objectType")

        for points, address in retrieve_type(objList, obj_type):
            request.append(f"{points} {address} {prop_list} ")

        def _find_propid_index(key):
            self.log(f"Prop List : {prop_list}", level="debug")
            _prop_list = prop_list.split(" ")
            for i, each in enumerate(_prop_list):
                if key == each:
                    return i
            raise KeyError(f"{key} not part of property list")

        try:
            self.log(f"Request : {request}", level="debug")
            points_info = await self.read_multiple(
                "",
                discover_request=(request, len(prop_list.split(" "))),
                points_per_request=points_per_request,
            )
            self.log(f"Points Info : {points_info}", level="debug")
        except SegmentationNotSupported:
            raise
        # Process responses and create point
        i = 0
        for each in retrieve_type(objList, obj_type):
            point_type = str(each[0])
            point_address = str(each[1])
            point_infos = points_info[i]
            i += 1
            self._log.debug(
                f"Retrieved Type {point_type} {point_address} {point_infos}"
            )
            pointName = point_infos[_find_propid_index("objectName")]
            presentValue = point_infos[_find_propid_index("presentValue")]
            self._log.debug(
                f"Reading {pointName} gave {presentValue} of type {obj_type}"
            )
            if presentValue is not None:
                if obj_type == "analog" or obj_type == "loop":
                    presentValue = float(presentValue)
                elif obj_type == "multi":
                    presentValue = int(presentValue)
            try:
                point_description = str(point_infos[_find_propid_index("description")])
            except KeyError:
                point_description = ""
            try:
                point_units_state = point_infos[_find_propid_index("units")]
            except KeyError:
                try:
                    point_units_state = point_infos[_find_propid_index("stateText")]
                except KeyError:
                    try:
                        _inactive = point_infos[_find_propid_index("inactiveText")]
                        _active = point_infos[_find_propid_index("activeText")]
                        point_units_state = (_inactive, _active)
                    except KeyError:
                        if obj_type == "binary":
                            point_units_state = ("OFF", "ON")
                        elif obj_type == "multi":
                            point_units_state = [""]
                        else:
                            point_units_state = None

            try:
                new_points.append(
                    obj_cls(
                        pointType=point_type,
                        pointAddress=point_address,
                        pointName=pointName,
                        description=point_description,
                        presentValue=presentValue,
                        units_state=point_units_state,
                        device=self,
                        history_size=self.properties.history_size,
                    )
                )
            except IndexError:
                self._log.warning(
                    "There has been a problem defining {} points. It is sometimes due to busy network. Please retry the device creation".format(
                        obj_type
                    )
                )
                raise
        return new_points


class RPObjectsProcessing:
    async def _process_new_objects(
        self, obj_cls=NumericPoint, obj_type: str = "analog", objList=None
    ):
        _newpoints = []
        for each in retrieve_type(objList, obj_type):
            point_type = str(each[0])
            point_address = str(each[1])

            if obj_type == "analog":
                units_state = await self.read_single(
                    f"{point_type} {point_address} units "
                )
            elif obj_type == "multi":
                units_state = await self.read_single(
                    f"{point_type} {point_address} stateText "
                )
            elif obj_type == "loop":
                units_state = await self.read_single(
                    f"{point_type} {point_address} units "
                )
            elif obj_type == "binary":
                units_state = (
                    (
                        await self.read_single(
                            f"{point_type} {point_address} inactiveText "
                        )
                    ),
                    (
                        await self.read_single(
                            f"{point_type} {point_address} activeText "
                        )
                    ),
                )
            else:
                units_state = None

            presentValue = await self.read_single(
                f"{point_type} {point_address} presentValue "
            )
            if (obj_type == "analog" or obj_type == "loop") and presentValue:
                presentValue = float(presentValue)

            _newpoints.append(
                obj_cls(
                    pointType=point_type,
                    pointAddress=point_address,
                    pointName=await self.read_single(
                        f"{point_type} {point_address} objectName "
                    ),
                    description=await self.read_single(
                        f"{point_type} {point_address} description "
                    ),
                    presentValue=presentValue,
                    units_state=units_state,
                    device=self,
                )
            )
        return _newpoints


class ReadPropertyMultiple(ReadUtilsMixin, DiscoveryUtilsMixin, RPMObjectsProcessing):
    async def read_multiple(
        self,
        points_list,
        *,
        points_per_request=25,
        discover_request=(None, 6),
        force_single=False,
        property_identifier="presentValue",
    ):
        """
        Read points from a device using a ReadPropertyMultiple request.
        [ReadProperty requests are very slow in comparison].

        :param points_list: (list) a list of all point_name as str
        :param points_per_request: (int) number of points in the request

        Requesting many points results big requests that need segmentation.  Aim to request
        just the 'right amount' so segmentation can be avoided.  Determining the 'right amount'
        is often trial-&-error.

        :Example:

        device.read_multiple(['point1', 'point2', 'point3'], points_per_request = 10)
        """
        if not self.properties.pss["readPropertyMultiple"] or force_single:
            self.log("Read property Multiple Not supported", level="warning")
            await self.read_single(
                points_list, points_per_request=1, discover_request=discover_request
            )
        else:
            if not self.properties.segmentation_supported:
                points_per_request = 1

            if discover_request[0]:
                values = []
                info_length = discover_request[1]
                big_request = discover_request[0]
                self.log(f"Discover : {big_request}", level="debug")
                self.log(f"Length : {info_length}", level="debug")

                for request in batch_requests(big_request, points_per_request):
                    try:
                        self.properties.address
                        request = f"{self.properties.address} {''.join(request)}"
                        self.log(f"RPM_Request: {request} ", level="debug")
                        try:
                            val = await self.properties.network.readMultiple(
                                request, vendor_id=self.properties.vendor_id
                            )
                        except SegmentationNotSupported:
                            raise
                        except ValueError as error:
                            # high limit ?
                            self._log.warning(
                                f"Got a value error of {error} for request : {request}"
                            )
                            self._log.warning(
                                "We will use single point reading to create device and turn off segmentation support"
                            )
                            self.properties.segmentation_supported = False
                            await self.read_multiple(
                                points_list,
                                points_per_request=1,
                                discover_request=discover_request,
                            )
                            break

                        # print('val : ', val, len(val), type(val))
                        if val is None:
                            self.properties.segmentation_supported = False
                            raise SegmentationNotSupported

                    except KeyError as error:
                        raise Exception(f"Unknown point name : {error}")

                    except SegmentationNotSupported:
                        self.properties.segmentation_supported = False
                        self._log.warning(
                            "Looks like segmentation is not supported. Turning that off."
                        )
                        # self.read_multiple(points_list,points_per_request=1, discover_request=discover_request)
                        self.log("Segmentation not supported", level="warning")
                        self.log("Request too big...will reduce it", level="warning")
                        if points_per_request == 1:
                            raise
                        await self.read_multiple(
                            points_list,
                            points_per_request=1,
                            discover_request=discover_request,
                        )

                    else:
                        for points_info in batch_requests(val, info_length):
                            values.append(points_info)
                return values

            else:
                self.log("Read Multiple", level="debug")
                big_request = self._rpm_request_by_name(
                    points_list, property_identifier=property_identifier
                )
                i = 0
                for request in batch_requests(big_request[0], points_per_request):
                    try:
                        request = f"{self.properties.address} {''.join(request)}"
                        self.log(request, level="debug")
                        val = await self.properties.network.readMultiple(
                            request, vendor_id=self.properties.vendor_id
                        )

                    except SegmentationNotSupported:
                        self.properties.segmentation_supported = False
                        await self.read_multiple(
                            points_list,
                            points_per_request=1,
                            discover_request=discover_request,
                        )

                    except KeyError as error:
                        raise Exception(f"Unknown point name : {error}")

                    else:
                        points_values = zip(big_request[1][i : i + len(val)], val)
                        i += len(val)
                        for each in points_values:
                            each[0]._trend(each[1])

    async def read_single(
        self, points_list, *, points_per_request=1, discover_request=(None, 4)
    ):
        if discover_request[0]:
            return await self.rp_discovered_values(
                discover_request, points_per_request=points_per_request
            )

        else:
            big_request = self._rpm_request_by_name(points_list)
            i = 0
            for request in batch_requests(big_request[0], points_per_request):
                try:
                    request = f"{self.properties.address} {''.join(request)}"
                    val = await self.properties.network.read(
                        request, vendor_id=self.properties.vendor_id
                    )
                    points_values = zip(big_request[1][i : i + len(val)], val)
                    i += len(val)
                    for each in points_values:
                        each[0]._trend(each[1])

                except KeyError as error:
                    raise Exception(f"Unknown point name : {error}")

    def poll(self, command="start", *, delay=10):
        """
        Poll a point every x seconds (delay=x sec)
        Can be stopped by using point.poll('stop') or .poll(0) or .poll(False)
        or by setting a delay = 0

        :param command: (str) start or stop polling
        :param delay: (int) time delay between polls in seconds
        :type command: str
        :type delay: int

        :Example:

        device.poll()
        device.poll('stop')
        device.poll(delay = 5)
        """
        _poll_cls: t.Union[t.Type[DeviceFastPoll], t.Type[DeviceNormalPoll]]
        if delay < 10:
            self.properties.fast_polling = True
            _poll_cls = DeviceFastPoll
        else:
            self.properties.fast_polling = False
            _poll_cls = DeviceNormalPoll

        if str(command).lower() not in ["stop", "start", "0", "False"]:
            self.log(
                'Bad argument for function. Needs "stop", "start", "0" or "False" or provide keyword arg (command or delay)',
                level="error",
            )
            return

        if (
            str(command).lower() == "stop"
            or command == False  # noqa E712
            or command == 0
            or delay == 0
        ):
            if isinstance(self._polling_task.task, DeviceNormalPoll) or isinstance(
                self._polling_task.task, DeviceFastPoll
            ):
                self._polling_task.task.stop()
                while self._polling_task.task.is_alive():
                    pass

                self._polling_task.task = None
                self._polling_task.running = False
                self.log(f"{self.properties.name} | Polling stopped", level="info")

        elif self._polling_task.task is None:
            self._polling_task.task = _poll_cls(
                self, delay=delay, name=self.properties.name
            )
            self._polling_task.task.start()
            self._polling_task.running = True
            self.log(
                f"Polling started, values read every {delay} seconds", level="info"
            )

        elif self._polling_task.running:
            self._polling_task.task.stop()
            while self._polling_task.task.is_alive():
                pass
            self._polling_task.running = False
            self._polling_task.task = _poll_cls(
                self, delay=delay, name=self.properties.name
            )
            self._polling_task.task.start()
            self._polling_task.running = True
            self.log(
                f"Polling started, every values read each {delay} seconds", level="info"
            )

        else:
            raise RuntimeError("Stop polling before redefining it")


class ReadProperty(ReadUtilsMixin, DiscoveryUtilsMixin, RPObjectsProcessing):
    async def read_multiple(
        self, points_list, *, points_per_request=1, discover_request=(None, 6)
    ):
        """
        Functions to read points from a device using the ReadPropertyMultiple request.
        Using readProperty request can be very slow to read a lot of data.

        :param points_list: (list) a list of all point_name as str
        :param points_per_request: (int) number of points in the request

        Using too many points will create big requests needing segmentation.
        It's better to use just enough request so the message will not require
        segmentation.

        :Example:

        device.read_multiple(['point1', 'point2', 'point3'], points_per_request = 10)
        """
        if isinstance(points_list, list):
            (requests, points) = self._rpm_request_by_name(points_list)
            for i, req in enumerate(requests):
                val = await self.read_single(
                    req, points_per_request=1, discover_request=discover_request
                )
                if val is not None and val != "":
                    points[i]._trend(val)
        else:
            await self.read_single(
                points_list, points_per_request=1, discover_request=discover_request
            )

    async def read_single(
        self, request, *, points_per_request=1, discover_request=(None, 4)
    ):
        try:
            request = f"{self.properties.address} {''.join(request)}"
            self.log(f"RP_Request: {request} ", level="debug")
            return await self.properties.network.read(
                request, vendor_id=self.properties.vendor_id
            )
        except KeyError as error:
            raise Exception(f"Unknown point name: {error}")

        except NoResponseFromController:
            return ""

    def poll(self, command="start", *, delay=120):
        """
        Poll a point every x seconds (delay=x sec)
        Can be stopped by using point.poll('stop') or .poll(0) or .poll(False)
        or by setting a delay = 0

        :param command: (str) start or stop polling
        :param delay: (int) time delay between polls in seconds
        :type command: str
        :type delay: int

        :Example:

        device.poll()
        device.poll('stop')
        device.poll(delay = 5)
        """
        if delay < 10:
            self._log.warning(
                "Device do not support RPM, fast polling not available, limiting delay to 10sec."
            )
            self.properties.fast_polling = False
            delay = 10

        if (
            str(command).lower() == "stop"
            or command is False
            or command == 0
            or delay == 0
        ):
            if isinstance(self._polling_task.task, DeviceNormalPoll):
                self._polling_task.task.stop()
                while self._polling_task.task.is_alive():
                    pass

                self._polling_task.task = None
                self._polling_task.running = False
                self.log("Polling stopped", level="info")

        elif self._polling_task.task is None:
            self._polling_task.task = DeviceNormalPoll(
                self, delay=delay, name=self.properties.name
            )
            self._polling_task.task.start()
            self._polling_task.running = True
            self.log(
                f"Polling started, values read every {delay} seconds", level="info"
            )

        elif self._polling_task.running:
            self._polling_task.task.stop()
            while self._polling_task.task.is_alive():
                pass
            self._polling_task.running = False
            self._polling_task.task = DeviceNormalPoll(
                self, delay=delay, name=self.properties.name
            )
            self._polling_task.task.start()
            self._polling_task.running = True
            self.log(
                f"Polling started, every values read each {delay} seconds", level="info"
            )

        else:
            raise RuntimeError("Stop polling before redefining it")



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#

# --- standard Python modules ---
# --- 3rd party modules ---

from ..core.devices.Device import Device
from ..core.io.IOExceptions import BadDeviceDefinition

# --- this application's modules ---
from .TaskManager import Task

"""
A way to define a BAC0.device using a task, so it won't block the Notebook or the REPL
TODO : check if still required, maybe deprecated
"""


class AddDevice(Task):
    def __init__(self, address=None, boid=None, network=None, callback=None, **kwargs):
        if network is None:
            raise ValueError("Please provide network")
        else:
            self.network = network
        self._kwargs = kwargs

        if address is not None and boid is not None:
            self.address = address
            self.boid = boid
        else:
            self.address, self.boid = self.find_address(address, boid)

        name = f"Adding_Device_{self.address}|{self.boid}"
        self.callback = callback
        self._kwargs["address"] = self.address
        self._kwargs["device_id"] = self.boid
        self._kwargs["network"] = network
        super().__init__(name=name, delay=0)
        # self.start()

    def find_address(self, address, boid):
        if self.network.discoveredDevices is None:
            self.log(
                "Device cannot be created yet, use bacnet.discover() or provide both address and Boid",
                level="error",
            )
            raise DeviceNotFoundError(
                "Device cannot be created yet, use bacnet.discover() or provide both address and Boid"
            )
        else:
            for each in self.network.discoveredDevices:
                address, boid = each
                if address in str(address) or boid in str(boid):
                    self.log(f"Found {each}", level="info")
                    return each
        self.log(
            "Device not discovered yet, use bacnet.discover() or provide both address and Boid",
            level="error",
        )

    def task(self, **kwargs):
        try:
            self.log(kwargs, level="info")
            dev = Device(kwargs)
            self.log(
                f"Device named {dev.properties.name} ({dev.properties.address}/{dev.properties.device_id}) created. Retrieve it using bacnet[boid]",
                level="info",
            )
            if self.callback is not None:
                self.log(f"Executing callback for {dev.properties.name}", level="info")
                self.callback()
        except BadDeviceDefinition as error:
            self.log(f"Bad device definition ({error})", level="error")


class DeviceNotFoundError(Exception):
    pass



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
TimeSync.py - creation of time synch requests

"""

import datetime as dt
from datetime import datetime

# --- standard Python modules ---

import pytz
from bacpypes3.apdu import TimeSynchronizationRequest, UTCTimeSynchronizationRequest
from bacpypes3.app import Application
from bacpypes3.basetypes import DateTime

# --- 3rd party modules ---
from bacpypes3.pdu import Address, GlobalBroadcast, LocalBroadcast
from bacpypes3.primitivedata import Date, Time

from BAC0.core.app.asyncApp import BAC0Application

from ...core.utils.notes import note_and_log
from ..io.IOExceptions import ApplicationNotStarted


def _build_datetime(UTC=False):
    if UTC:
        _d = dt.datetime.utcnow().date()
        _t = dt.datetime.utcnow().time()
        _date = Date((_d.year - 1900, _d.month, _d.day, _d.isoweekday()))
        _time = Time(
            (
                _t.hour,
                _t.minute,
                _t.second,
                int(_t.microsecond / 10000),
            )
        )
    else:
        _date = Date().now()
        _time = Time().now()
    return DateTime(date=_date, time=_time)


@note_and_log
class TimeSync:
    """
    Mixin to support Time Synchronisation from BAC0 to other devices
    """

    def time_sync(
        self, destination: str = None, datetime: DateTime = None, UTC: bool = False
    ) -> None:
        """
        Take local time and send it to devices. User can also provide
        a datetime value (constructed following bacpypes.basetypes.Datetime
        format).

        To create a DateTime ::

            from bacpypes.basetypes import DateTime
            from bacpypes.primitivedata import Date, Time

            # Create date and time
            _date = Date('2019-08-05')
            _time = Time('16:45')

            # Create Datetime
            _datetime = DateTime(date=_date.value, time=_time.value)

            # Pass this to the function
            bacnet.time_sync(datetime=_datetime)


        """
        if not self._started:
            raise ApplicationNotStarted("BACnet stack not running - use startApp()")

        _this_application: BAC0Application = self.this_application
        _app: Application = _this_application.app

        if not datetime:
            _datetime = _build_datetime(UTC=UTC)
        elif isinstance(datetime, DateTime):
            _datetime = datetime
        else:
            raise ValueError(
                "Please provide valid DateTime in bacpypes.basetypes.DateTime format"
            )

        # build a request
        if UTC:
            request = UTCTimeSynchronizationRequest(time=_datetime)
        else:
            request = TimeSynchronizationRequest(time=_datetime)

        if destination:
            if destination.lower() == "global":
                request.pduDestination = GlobalBroadcast()
            elif destination.lower() == "local":
                request.pduDestination = LocalBroadcast()
            else:
                try:
                    request.pduDestination = Address(destination)
                except (TypeError, ValueError):
                    self.log(
                        "Destination unrecognized ({destination}), setting local broadcast",
                        level="warning",
                    )
                    request.pduDestination = LocalBroadcast()
        else:
            request.pduDestination = LocalBroadcast()

        self.log(f"{'- request:':>12} {request}", level="debug")

        _app.request(request)

        year, month, day, dow = _datetime.date
        year = year + 1900
        hour, minutes, sec, msec = _datetime.time
        d = dt.datetime(year, month, day, hour, minutes, sec, msec)
        self.log(f"Time Sync Request sent to network : {d.isoformat()}", level="info")


class TimeHandler(object):
    """
    This class will deal with Time / Timezone related features
    To deal with DateTime Value correctly we need to be aware
    of timezone.
    """

    def __init__(self, tz: str = "America/Montreal") -> None:
        self.set_timezone(tz)

    def set_timezone(self, tz: str) -> None:
        self.timezone = pytz.timezone(tz)

    @property
    def now(self) -> datetime:
        return dt.datetime.now()

    def local_time(self):
        return self.now.time()

    def local_date(self) -> datetime.date:
        return self.now.date()

    def utcOffset(self) -> float:
        "Returns UTC offset in minutes"
        return round(self.now.astimezone().utcoffset().total_seconds() / 60)  # type: ignore[union-attr]

    def is_dst(self) -> bool:
        return self.timezone.dst(self.now) != dt.timedelta(0)

    def __repr__(self):
        return f"{self.__dict__}"



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
DoOnce.py - execute a task once
"""

import asyncio
from ..core.utils.notes import note_and_log
from .TaskManager import OneShotTask


@note_and_log
class DoOnce(OneShotTask):
    """
    Start a polling task which is in fact a recurring read of the point.
    ex.
        device['point_name'].poll(delay=60)
    """

    def __init__(self, fnc, name="do_once"):
        """
        :param point: (BAC0.core.device.Points.Point) name of the point to read
        :param delay: (int) Delay between reads in seconds, defaults = 10sec

        A delay cannot be < 5sec (there are risks of overloading the device)

        :returns: Nothing
        """
        super().__init__(fnc, name=name)

    async def task(self):
        if self.fnc_args:
            if asyncio.iscoroutinefunction(self.func):
                self._log.debug(
                    f"Running {self.func.__name__} with args {self.fnc_args}"
                )
                await self.func(self.fnc_args)
            else:
                self.func(self.fnc_args)
        else:
            if asyncio.iscoroutinefunction(self.func):
                self.log(f"Running {self.func.__name__}", level="debug")
                await self.func()
            else:
                self.func()



from bacpypes3.apdu import WritePropertyRequest
from bacpypes3.app import Application
from bacpypes3.constructeddata import Any
from bacpypes3.pdu import Address
from bacpypes3.primitivedata import CharacterString

from BAC0.core.app.asyncApp import BAC0Application


class TextMixin:
    """
    Mixin with functions to deal with text properties.
    Adding features to "network" itself.
    """

    def send_text_write_request(
        self,
        addr: str,
        obj_type: str,
        obj_inst: int,
        value: str,
        prop_id: str = "description",
    ) -> None:
        request = self.build_text_write_request(
            addr=addr,
            obj_type=obj_type,
            obj_inst=obj_inst,
            value=value,
            prop_id=prop_id,
        )
        self.write_text_value(request)

    def build_text_write_request(
        self,
        addr: str,
        obj_type: str,
        obj_inst: int,
        value: str,
        prop_id: str = "description",
    ) -> WritePropertyRequest:
        request = WritePropertyRequest(
            objectIdentifier=(obj_type, obj_inst), propertyIdentifier=prop_id
        )
        request.pduDestination = Address(addr)

        _value = Any(None)
        _value.cast_in(CharacterString(value))
        request.propertyValue = _value

        return request

    def write_text_value(
        self, request: WritePropertyRequest, timeout: int = 10
    ) -> None:
        _this_application: BAC0Application = self.this_application
        _app: Application = _this_application.app

        self.log(f"{'- request:':>12} {request}", level="debug")

        _app.request(request)



"""
#   PRIVA BV 105
#   Proprietary Objects and their attributes
#   reference needed
#
# Register Priva BV Proprietary Objects and properties
"""

# from . import OptionalUnsigned
from bacpypes3.basetypes import OptionalUnsigned, PropertyIdentifier
from bacpypes3.debugging import ModuleLogger
from bacpypes3.object import AnalogInputObject as _AnalogInputObject
from bacpypes3.object import AnalogOutputObject as _AnalogOutputObject
from bacpypes3.object import AnalogValueObject as _AnalogValueObject
from bacpypes3.object import DeviceObject as _DeviceObject
from bacpypes3.object import NetworkPortObject as _NetworkPortObject
from bacpypes3.primitivedata import Boolean  # Signed,
from bacpypes3.primitivedata import (
    Atomic,
    CharacterString,
    Date,
    Enumerated,
    ObjectType,
    Real,
    Time,
    Unsigned,
)
from bacpypes3.vendor import VendorInfo, get_vendor_info

# some debugging
_debug = 0
_log = ModuleLogger(globals())


# this vendor identifier reference is used when registering custom classes
_vendor_id = 105
_vendor_name = "PRIVA BV"


class ProprietaryObjectType(ObjectType):
    """
    This is a list of the object type enumerations for proprietary object types,
    see Clause 23.4.1.
    """

    pass


class ProprietaryPropertyIdentifier(PropertyIdentifier):
    """
    This is a list of the property identifiers that are used in custom object
    types or are used in custom properties of standard types.
    """

    custom_description = 10502


try:
    _priva = VendorInfo(
        _vendor_id, ProprietaryObjectType, ProprietaryPropertyIdentifier
    )
except RuntimeError:
    pass  # we are re-running the script... forgive us or maybe we already read a priva bv device
    _priva = get_vendor_info(_vendor_id)


class PrivaBVDeviceObject(_DeviceObject):
    """
    When running as an instance of this custom device, the DeviceObject is
    an extension of the one defined in bacpypes3.local.device
    """

    custom_description: CharacterString


class NetworkPortObject(_NetworkPortObject):
    """
    When running as an instance of this custom device, the NetworkPortObject is
    an extension of the one defined in bacpypes3.local.networkport (in this
    case doesn't add any proprietary properties).
    """

    pass


class PrivaBVAnalogInputObject(_AnalogInputObject):
    custom_description: CharacterString


class PrivaBVAnalogValueObject(_AnalogValueObject):
    custom_description: CharacterString


class PrivaBVAnalogOutputObject(_AnalogOutputObject):
    custom_description: CharacterString



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
RecurringTask.py - execute a recurring task
"""
import asyncio
from typing import Any, Callable, Tuple, Union, Coroutine
from concurrent.futures import ThreadPoolExecutor

from ..core.utils.notes import note_and_log
from .TaskManager import Task


@note_and_log
class RecurringTask(Task):
    """
    Start a recurring task (a function passed)
    """

    def __init__(
        self,
        fnc: Union[Tuple[Callable, Any], Callable, Coroutine],
        delay: int = 60,
        name: str = "recurring",
    ) -> None:
        """
        :param fnc: a function or a tuple (function, args)
        :param delay: (int) Delay between reads executions

        :returns: Nothing
        """
        self.fnc_args = None
        self.delay = delay
        if isinstance(fnc, tuple):
            self.func, self.fnc_args = fnc
        elif hasattr(fnc, "__call__"):
            self.func = fnc
        else:
            raise ValueError(
                "You must pass a function or a tuple (function,args) to this..."
            )
        Task.__init__(self, name=name, delay=delay)

    async def task(self) -> None:
        loop = asyncio.get_event_loop()
        executor = ThreadPoolExecutor()
        if self.fnc_args:
            if asyncio.iscoroutinefunction(self.func):
                await self.func(self.fnc_args)
            else:
                await loop.run_in_executor(executor, self.func, self.fnc_args)
                #self.func(self.fnc_args)
        else:
            if asyncio.iscoroutinefunction(self.func):
                await self.func()
            else:
                await loop.run_in_executor(executor, self.func)
                #self.func()
        await asyncio.sleep(self.delay)



import asyncio
from typing import Tuple

from bacpypes3.basetypes import BinaryPV, PropertyIdentifier
from bacpypes3.pdu import Address
from bacpypes3.primitivedata import Boolean, CharacterString, ObjectIdentifier

from ...core.app.asyncApp import BAC0Application
from ...scripts.Base import Base


class COVSubscription:
    def __init__(
        self,
        address: Address = None,
        objectID: Tuple[str, int] = None,
        lifetime: int = 900,
        confirmed: bool = False,
        callback=None,
        BAC0App: BAC0Application = None,
    ):
        self.address = Address(address) if isinstance(address, str) else address
        self.cov_fini = asyncio.Event()
        self.task = None
        self.obj_identifier = ObjectIdentifier(objectID)
        self._lite = BAC0App
        self._this_application = self._lite.this_application
        self._app = self._this_application.app
        self.callback = callback
        self.process_identifier = Base._last_cov_identifier + 1
        Base._last_cov_identifier = self.process_identifier

        # self.point = point
        self.lifetime = lifetime
        self.confirmed = confirmed
        self.callback = callback

    async def run(self):
        self._lite._log.debug(
            f"Subscribing to COV for {self.address} | {self.obj_identifier} -> {self._app}"
        )

        try:
            async with self._app.change_of_value(
                self.address,
                self.obj_identifier,
                self.process_identifier,
                self.confirmed,
                self.lifetime,
            ) as scm:
                cov_fini_task_monitor = asyncio.create_task(self.cov_fini.wait())
                while not self.cov_fini.is_set():
                    incoming: asyncio.Future = asyncio.ensure_future(scm.get_value())
                    done, pending = await asyncio.wait(
                        [
                            incoming,
                        ],
                        return_when=asyncio.FIRST_COMPLETED,
                    )
                    for task in pending:
                        task.cancel()

                    if incoming in done:
                        property_identifier, property_value = incoming.result()
                        self._lite._log.debug(
                            f"COV notification received for {self.address}|{self.obj_identifier} | {property_identifier} -> {type(property_identifier)} with value {property_value} | {property_value} -> {type(property_value)}"
                        )
                        if self.callback is not None:
                            self._lite._log.debug(
                                f"Calling callback for {property_identifier}"
                            )
                            if asyncio.iscoroutinefunction(self.callback):
                                asyncio.create_task(
                                    self.callback(
                                        property_identifier=property_identifier,
                                        property_value=property_value,
                                    )
                                )
                            elif hasattr(self.callback, "__call__"):
                                self.callback(
                                    property_identifier=property_identifier,
                                    property_value=property_value,
                                )
                            else:
                                self._lite._log.error(
                                    f"Callback {self.callback} is not callable"
                                )
                await cov_fini_task_monitor
        except Exception as e:
            self._lite._log.error(f"Error in COV subscription : {e}")

    def stop(self):
        self._lite._log.debug(
            f"Stopping COV subscription class for {self.address} | {self.obj_identifier}"
        )
        self.cov_fini.set()



from bacpypes3.local.object import Object as _Object

from bacpypes3.object import CharacterStringValueObject as _CharacterStringValueObject
from bacpypes3.object import DateTimeValueObject as _DateTimeValueObject
from bacpypes3.object import DateValueObject as _DateValueObject
from bacpypes3.object import TrendLogObject as _TrendLogObject

"""
Those models are opiniated versions of BACnet objects.
They are meant to supply an easy mechanism to build BACnet objects
that will be served by the local device of BAC0.

You can indeed use this "method" and modify it to fill your needs.
"""


# First create our local classes which are missing from bacpypes3


class CharacterStringValueObject(_Object, _CharacterStringValueObject):
    pass


class DateValueObject(_Object, _DateValueObject):
    pass


class DateTimeValueObject(_Object, _DateTimeValueObject):
    pass


class TrendLogObject(_Object, _TrendLogObject):
    pass



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
infos.py - BAC0 Package MetaData
"""

__author__ = "Christian Tremblay, P.Eng."
__email__ = "christian.tremblay@servisys.com"
__url__ = "https://github.com/ChristianTremblay/BAC0"
__download_url__ = "https://github.com/ChristianTremblay/BAC0/archive/master.zip"
__version__ = "2024.09.10"
__license__ = "LGPLv3"



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
Poll.py - create a Polling task to repeatedly read a point.
"""

import typing as t

# --- standard Python modules ---
import weakref

from ..core.utils.notes import note_and_log

# --- this application's modules ---
from .TaskManager import Task

if t.TYPE_CHECKING:
    from ..core.devices.Device import RPDeviceConnected, RPMDeviceConnected


# ------------------------------------------------------------------------------
class MultiplePollingFailures(Exception):
    pass


@note_and_log
class SimplePoll(Task):
    """
    Start a polling task to repeatedly read a point's Present_Value.
    ex.
        device['point_name'].poll(delay=60)
    """

    def __init__(self, point, *, delay: int = 10) -> None:
        """
        :param point: (BAC0.core.device.Points.Point) name of the point to read
        :param delay: (int) Delay between reads in seconds, defaults = 10sec

        A delay cannot be < 1sec
        This task is meant for single points, so BAC0 will allow short delays.
        This way, a fast polling is available for some points in a device that
        would not support segmentation.

        :returns: Nothing
        """
        if delay < 1:
            delay = 1
        if point.properties:
            self._point = point
            Task.__init__(self, name="rp_poll", delay=delay)
        else:
            raise ValueError("Provide a point object")

    async def task(self):
        await self._point.value


@note_and_log
class DevicePoll(Task):
    """
    Start a polling task to repeatedly read a list of points from a device using
    ReadPropertyMultiple requests.
    """

    def __init__(
        self,
        device: t.Union["RPMDeviceConnected", "RPDeviceConnected"],
        delay: int = 10,
        name: str = "",
        prefix: str = "basic_poll",
    ) -> None:
        """
        :param device: (BAC0.core.devices.Device.Device) device to poll
        :param delay: (int) Delay between polls in seconds, defaults = 10sec

        A delay cannot be < 10sec
        For delays under 10s, use DeviceFastPoll class.

        :returns: Nothing
        """
        self.failures = 0
        self.MAX_FAILURES = 3
        self._device = weakref.ref(device)
        Task.__init__(self, name=f"{prefix}_{name}", delay=delay)
        self._counter = 0

    @property
    def device(self) -> t.Union["RPMDeviceConnected", "RPDeviceConnected", None]:
        return self._device()

    async def task(self) -> None:
        if self.device.properties.ping_failures > 0:
            self.device._log.warning(
                "{} ({}) | Ping failed, skipping polling for now. Resending a ping to speed up things".format(
                    self.device.properties.name, self.device.properties.address
                )
            )
            await self.device.ping()
            return
        try:
            if self.failures >= self.MAX_FAILURES:
                raise MultiplePollingFailures(
                    "{} ({}) | Polling failed numerous times in a row... let see what we can do".format(
                        self.device.properties.name, self.device.properties.address
                    )
                )
            await self.device.read_multiple(
                list(self.device.pollable_points_name), points_per_request=25
            )
            self._counter += 1
            if self._counter == self.device.properties.auto_save:
                self.device.save(resampling=self.device.properties.save_resampling)
                if self.device.properties.clear_history_on_save:
                    self.device.clear_histories()
                self._counter = 0
            self.failures = 0
        except AttributeError as e:
            # This error can be seen when defining a controller on a busy network...
            # When creation fail, polling is created and fail the first time...
            # So kill the task
            self.device._log.error(
                "{} ({}) | Something is wrong while creating the polling task.\nError: {} | Type : {}".format(
                    self.device.properties.name,
                    self.device.properties.address,
                    e,
                    type(e),
                )
            )
            # self.stop()
            self.failures += 1
        except ValueError as e:
            self.failures += 1
            self.device._log.error(
                "{} ({}) | Polling results contains a wrong value. Probably a communication error. Will skip this result and wait for the next cycle.\nError: {} | Type : {}".format(
                    self.device.properties.name,
                    self.device.properties.address,
                    e,
                    type(e),
                )
            )
            pass

        except MultiplePollingFailures as e:
            self.device._log.warning(
                "{} ({}) | Trying to ping device then we'll reset the number of failures and get back with polling\nError: {}| Type : {}".format(
                    self.device.properties.name,
                    self.device.properties.address,
                    e,
                    type(e),
                )
            )
            if await self.device.ping():
                self.failures = 0


@note_and_log
class DeviceOneShotPoll(DevicePoll):
    """
    Start a normal polling task to repeatedly read a list of points from a device using
    ReadPropertyMultiple requests.

    Normal polling will limit the polling speed to 10 second minimum

    """

    def __init__(self, device, name=""):
        """
        :param device: (BAC0.core.devices.Device.Device) device to poll
        :param delay: (int) Delay between polls in seconds, defaults = 10sec

        :returns: Nothing
        """
        self._log.info(f"Starting a simple polling task for {device.properties.name}")
        DevicePoll.__init__(
            self, device=device, name=name, delay=0, prefix="rpm_oneshot_poll"
        )


@note_and_log
class DeviceNormalPoll(DevicePoll):
    """
    Start a normal polling task to repeatedly read a list of points from a device using
    ReadPropertyMultiple requests.

    Normal polling will limit the polling speed to 10 second minimum

    """

    def __init__(self, device, delay=10, name=""):
        """
        :param device: (BAC0.core.devices.Device.Device) device to poll
        :param delay: (int) Delay between polls in seconds, defaults = 10sec

        :returns: Nothing
        """
        if delay < 10:
            delay = 10
        self._log.info(f"Device defined for normal polling with a delay of {delay}sec")
        DevicePoll.__init__(
            self, device=device, name=name, delay=delay, prefix="rpm_normal_poll"
        )


@note_and_log
class DeviceFastPoll(DevicePoll):
    """
    Start a fast polling task to repeatedly read a list of points from a device using
    ReadPropertyMultiple requests.
    Delay allowed will be 0 to 10 seconds
    Normal polling will limit the polling speed to 10 second minimum

    Warning : Fast polling must be used with care or network flooding may occur

    """

    def __init__(self, device, delay=1, name=""):
        """
        :param device: (BAC0.core.devices.Device.Device) device to poll
        :param delay: (int) Delay between polls in seconds, defaults = 1sec

        :returns: Nothing
        """
        if delay < 0:
            delay = 0.01
        elif delay > 10:
            delay = 10
        self._log.warning(f"Device defined for fast polling with a delay of {delay}sec")
        DevicePoll.__init__(
            self, device=device, name=name, delay=delay, prefix="rpm_fast_poll"
        )



"""
#   JCI 5
#   Proprietary Objects and their attributes
#   https://cgproducts.johnsoncontrols.com/MET_PDF/12013102.pdf
#
# Register Johnson Controls Proprietary Objects and properties
"""

from bacpypes3.vendor import VendorInfo, get_vendor_info
from bacpypes3.basetypes import PropertyIdentifier
from bacpypes3.debugging import ModuleLogger
from bacpypes3.object import AnalogInputObject as _AnalogInputObject
from bacpypes3.object import AnalogOutputObject as _AnalogOutputObject
from bacpypes3.object import AnalogValueObject as _AnalogValueObject
from bacpypes3.object import DeviceObject as _DeviceObject
from bacpypes3.object import NetworkPortObject as _NetworkPortObject
from bacpypes3.primitivedata import Boolean  # Signed,
from bacpypes3.primitivedata import (
    Atomic,
    CharacterString,
    Date,
    Enumerated,
    ObjectType,
    Real,
    Time,
    Unsigned,
)

# from . import OptionalUnsigned
from bacpypes3.basetypes import OptionalUnsigned

# some debugging
_debug = 0
_log = ModuleLogger(globals())


# this vendor identifier reference is used when registering custom classes
_vendor_id = 5
_vendor_name = "Johnson Controls"


class ProprietaryObjectType(ObjectType):
    """
    This is a list of the object type enumerations for proprietary object types,
    see Clause 23.4.1.
    """

    pass


class ProprietaryPropertyIdentifier(PropertyIdentifier):
    """
    This is a list of the property identifiers that are used in custom object
    types or are used in custom properties of standard types.
    """

    # DEVICE OBJECT
    alarm_state = 1006
    archive_date = 849
    archive_status = 1187
    archive_time = 850
    cpu_usage = 2583
    enabled = 673
    execution_priority = 2197
    extended_proto_version = 2291
    flash_usage = 2584
    item_reference = 32527
    jci_status = 847
    memory_usage = 2581
    object_category = 908
    object_memory_usage = 2582
    status = 512
    bacnet_broadcast_receive_rate = 745
    default_base_units = 2206
    estimated_flash_available = 2395
    last_idle_sample = 30082
    max_message_buffer = 848
    user_name = 2390
    pcode = 1320
    sab_device_status_list_changed = 4514
    events_lost = 1479
    accept_bacnet_time_sync = 4970
    supervisory_device_online = 3653
    # "STANDARD_TIME_OFFSET ": {"obj_id": 1017, "datatype": Signed, "mutable": False}, # -900-900
    # "DAYLIGHT_SAVING_TIME_OFFSET ": {"obj_id": 1093, "datatype": Signed, "mutable": False}, # -900-900
    # "STANDARD_TIME_START ": {"obj_id": 988, "datatype": Complex, "mutable": True},
    # "DAYLIGHT_SAVING_TIME_START ": {"obj_id": 1040, "datatype": Complex, "mutable": True},
    # "LAST_BACNET_TIME_SYNC_RECEIVED ": {"obj_id": 5728, "datatype": Complex, "mutable": False},
    # "ASSET_VERSIONS": {"obj_id": 4960, "datatype": Complex, "mutable": False},
    supervisory_offline_timeout = 6002
    next_available_oid = 787
    has_unbound_references = 767
    surrogate_cache_cnt = 571
    surrogate_cache_max = 639
    load_balancer_level = 4722
    timer_db_size = 733
    timer_used = 734
    timer_peak = 735
    timer_messages_aborted = 4959
    bacnet_oid_allocated = 1291
    bacnet_oid_used = 1292
    sign_pri_db_size = 730
    sign_pri_used = 731
    sign_pri_peak = 732
    bacnet_encode_type = 32578
    routing_mode = 4307
    bacnet_integrated_objects = 4302
    bacnet_compatible = 4581
    send_i_am_rate = 579
    default_time_zone = 32583
    time_zone = 1403
    local_time_zone = 1404
    fc_multicast_responder = 3390
    fc_wait_before_polling = 3391
    supervisor_mac_address = 3652
    library_part_id = 3295
    application_class_set_version = 4128
    device_model_class_set_version = 4129
    cov_min_send_time = 3929
    cov_transmits_per_minute = 651
    control_sequence_in_test = 3651
    end_of_line = 603
    device_address = 876
    fc_bus_communication_mod = 4400
    system_type = 3900
    system_configuration = 3899
    sabusperformance = 12157
    sabustokenlooptime = 12158
    sabuscovrcvperminute = 12159
    sabuscovwritesperminute = 12160
    # AV, AI, AO
    flow_sp_eeprom = 3113
    offset = 956
    offline = 913
    sabusaddr = 3645
    peertopeer = 748
    p2p_errorstatus = 746
    inputrangelow = 1293
    inputrangehigh = 1294
    outputrangelow = 1295
    outputrangehigh = 1296
    min_out_value = 652
    max_out_value = 653
    # polarity = polarity
    stroketime = 3478


try:
    _jci = VendorInfo(_vendor_id, ProprietaryObjectType, ProprietaryPropertyIdentifier)
except RuntimeError:
    pass  # we are re-running the script... forgive us or maybe we already read a jci device
    _jci = get_vendor_info(_vendor_id)


class JCIDeviceObject(_DeviceObject):
    """
    When running as an instance of this custom device, the DeviceObject is
    an extension of the one defined in bacpypes3.local.device
    """

    alarm_state: Enumerated
    archive_date: Date
    archive_status: Unsigned
    archive_time: Time
    cpu_usage: Real
    enabled: Boolean
    execution_priority: Enumerated
    extended_proto_version: Unsigned
    flash_usage: Real
    item_reference: CharacterString
    jci_status: Enumerated
    memory_usage: Real
    object_category: Enumerated
    object_memory_usage: Real
    status: Enumerated
    bacnet_broadcast_receive_rate: Unsigned
    default_base_units: Enumerated
    estimated_flash_available: Real
    last_idle_sample: Real
    max_message_buffer: Unsigned
    user_name: CharacterString
    pcode: CharacterString
    sab_device_status_list_changed: OptionalUnsigned
    events_lost: Unsigned
    accept_bacnet_time_sync: Boolean
    supervisory_device_online: Boolean
    supervisory_offline_timeout: Unsigned
    next_available_oid: Unsigned
    has_unbound_references: Boolean
    surrogate_cache_cnt: Unsigned
    surrogate_cache_max: Unsigned
    load_balancer_level: Enumerated
    timer_db_size: Unsigned
    timer_used: Unsigned
    timer_peak: Unsigned
    timer_messages_aborted: Unsigned
    bacnet_oid_allocated: Unsigned
    bacnet_oid_used: Unsigned
    sign_pri_db_size: Unsigned
    sign_pri_used: Unsigned
    sign_pri_peak: Unsigned
    bacnet_encode_type: Unsigned
    routing_mode: Enumerated
    bacnet_integrated_objects: Enumerated
    bacnet_compatible: Unsigned
    send_i_am_rate: Unsigned
    default_time_zone: Enumerated
    time_zone: Enumerated
    local_time_zone: Enumerated
    fc_multicast_responder: Unsigned
    fc_wait_before_polling: Unsigned
    supervisor_mac_address: Unsigned
    library_part_id: CharacterString
    application_class_set_version: Unsigned
    device_model_class_set_version: Unsigned
    cov_min_send_time: Unsigned
    cov_transmits_per_minute: Unsigned
    control_sequence_in_test: Unsigned
    end_of_line: Boolean
    device_address: Unsigned
    fc_bus_communication_mod: Enumerated
    system_type: Unsigned
    system_configuration: Unsigned
    sabusperformance: Unsigned
    sabustokenlooptime: Unsigned
    sabuscovrcvperminute: Unsigned
    sabuscovwritesperminute: Unsigned


class NetworkPortObject(_NetworkPortObject):
    """
    When running as an instance of this custom device, the NetworkPortObject is
    an extension of the one defined in bacpypes3.local.networkport (in this
    case doesn't add any proprietary properties).
    """

    pass


class JCIAnalogInputObject(_AnalogInputObject):
    offset: Real
    offline: Boolean
    sabusaddr: OptionalUnsigned
    inputrangelow: Real
    inputrangehigh: Real
    outputrangelow: Real
    outputrangehigh: Real


class JCIAnalogValueObject(_AnalogValueObject):
    flow_sp_eeprom: Real
    offset: Real
    offline: Boolean
    sabusaddr: OptionalUnsigned
    peertopeer: Atomic
    p2p_errorstatus: Enumerated


class JCIAnalogOutputObject(_AnalogOutputObject):
    offline: Boolean
    sabusaddr: OptionalUnsigned
    min_out_value: Real
    max_out_value: Real
    # polarity = polarity
    stroketime: Real



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
Doc here
"""
import asyncio
import random
import sys
import typing as t
from collections import defaultdict

# --- standard Python modules ---
from bacpypes3.basetypes import DeviceStatus, HostNPort, ObjectTypesSupported
from bacpypes3.json.util import sequence_to_json
from bacpypes3.local.device import DeviceObject
from bacpypes3.local.networkport import NetworkPortObject
from bacpypes3.pdu import Address
from bacpypes3.primitivedata import CharacterString
from bacpypes3.vendor import VendorInfo, get_vendor_info

# --- this application's modules ---
from ..core.app.asyncApp import (
    BAC0Application,
)  # BAC0BBMDDeviceApplication,; BAC0ForeignDeviceApplication,
from ..core.functions.GetIPAddr import validate_ip_address
from ..core.functions.TimeSync import TimeHandler
from ..core.io.IOExceptions import InitializationError, UnknownObjectError
from ..core.utils.notes import note_and_log
from ..tasks.TaskManager import stopAllTasks

# --- 3rd party modules ---

# ------------------------------------------------------------------------------


@note_and_log
class LocalObjects(object):
    def __init__(self, device):
        self.device = device

    def __getitem__(self, obj):
        item = None
        if isinstance(obj, tuple):
            obj_type, instance = obj
            item = self.device.this_application.app.get_object_id((obj_type, instance))
        elif isinstance(obj, str):
            name = obj
            item = self.device.this_application.app.get_object_name(name)
        if item is None:
            raise UnknownObjectError(f"Can't find {obj} in local device")
        else:
            return item


def charstring(val):
    return CharacterString(val) if isinstance(val, str) else val


@note_and_log
class Base:
    """
    Build a running BACnet/IP device that accepts WhoIs and IAm requests
    Initialization requires some minimial information about the local device.

    :param localIPAddr='127.0.0.1':
    :param localObjName='BAC0':
    :param deviceId=None:
    :param maxAPDULengthAccepted='1024':
    :param maxSegmentsAccepted='1024':
    :param segmentationSupported='segmentedBoth':
    """

    _used_ips: t.Set[Address] = set()
    _last_cov_identifier = 0
    _running_cov_tasks = {}

    def __init__(
        self,
        localIPAddr: Address = Address("127.0.0.1/24"),
        networkNumber: int = None,
        localObjName: str = "BAC0",
        deviceId: int = None,
        firmwareRevision: str = "".join(sys.version.split("|")[:2]),
        maxAPDULengthAccepted: str = "1024",
        maxSegmentsAccepted: str = "1024",
        segmentationSupported: str = "segmentedBoth",
        bbmdAddress: str = None,
        bbmdTTL: int = 0,
        bdtable: list = None,
        modelName: str = "BAC0 Scripting Tool",
        vendorId: int = 842,
        vendorName: str = "SERVISYS inc.",
        description: str = "http://christiantremblay.github.io/BAC0/",
        location: str = "Bromont, Québec",
        timezone: str = "America/Montreal",
        json_file: str = None,
    ):
        self.log("Configurating app", level="debug")

        # Register Servisys
        try:
            _BAC0_vendor = VendorInfo(vendorId)
        except RuntimeError:
            pass  # we are re-running the script... forgive us
            _BAC0_vendor = get_vendor_info(vendorId)
        _BAC0_vendor.register_object_class(
            ObjectTypesSupported.networkPort, NetworkPortObject
        )
        _BAC0_vendor.register_object_class(ObjectTypesSupported.device, DeviceObject)

        self.timehandler = TimeHandler(tz=timezone)

        self.response = None
        self._initialized = False
        self._started = False
        self._stopped = False

        if localIPAddr in Base._used_ips:
            raise InitializationError(
                "IP Address provided ({}) already used by BAC0. Check if another software is using port 47808 on this network interface. If so, you can define multiple IP per interface. Or specify another IP using BAC0.lite(ip='IP/mask')".format(
                    localIPAddr
                )
            )

        if validate_ip_address(localIPAddr):
            self.localIPAddr = localIPAddr
        else:
            raise InitializationError(
                "IP Address provided ({}) invalid. Check if another software is using port 47808 on this network interface. If so, you can define multiple IP per interface. Or specify another IP using BAC0.lite(ip='IP/mask')".format(
                    localIPAddr
                )
            )
        self.networkNumber = networkNumber

        self.Boid = (
            int(deviceId) if deviceId else (3056177 + int(random.uniform(0, 1000)))
        )

        self.segmentationSupported = segmentationSupported
        self.maxSegmentsAccepted = maxSegmentsAccepted
        self.localObjName = localObjName
        self.local_objects = LocalObjects(device=self)

        self.maxAPDULengthAccepted = maxAPDULengthAccepted
        self.vendorId = vendorId
        self.vendorName = charstring(vendorName)
        self.modelName = charstring(modelName)
        self.description = charstring(description)
        self.location = charstring(location)

        self.discoveredDevices: t.Optional[t.Dict[t.Tuple[str, int], int]] = None
        self.systemStatus = DeviceStatus(1)

        self.bbmdAddress = bbmdAddress
        self.bbmdTTL = bbmdTTL
        self.bdtable = bdtable

        self.firmwareRevision = firmwareRevision
        self._ric = {}
        self.subscription_contexts = {}
        self.database = None
        self.json_file = json_file

        try:
            self.startApp()
        except InitializationError as error:
            raise InitializationError(
                f"Gros probleme : {error}. Address requested : {localIPAddr}"
            )

    def startApp(self):
        """
        Define the local device, including services supported.
        Once defined, start the BACnet stack in its own thread.
        """
        self.log("Create Local Device", level="debug")
        try:
            app_type = "BACnet/IP App"

            class config(defaultdict):
                "Simple class to mimic args dot retrieval"

                def __init__(self, cfg):
                    for k, v in cfg.items():
                        self[k] = v

                def __getattr__(self, key):
                    return self[key]

            if self.bbmdAddress is not None:
                mode = "foreign"
            elif self.bdtable:
                mode = "bbmd"
            else:
                mode = "normal"
            cfg = {
                "BAC0": {
                    "bbmdAddress": self.bbmdAddress,
                    "bdt": self.bdtable,
                    "ttl": self.bbmdTTL,
                },
                "device": {
                    "object-name": self.localObjName,
                    # "firmware-revision": self.firmwareRevision,
                    "vendor-identifier": self.vendorId,
                    "vendor-name": "Servisys inc.",
                    "object-identifier": f"device,{self.Boid}",
                    "object-list": [f"device,{self.Boid}", "network-port,1"],
                    "model-name": self.modelName,
                    # "max-apdu-length-accepted": self.maxAPDULengthAccepted,
                    # "max-segments-accepted": self.maxSegmentsAccepted,
                    # "location": self.location,
                    # "description": self.description
                },
                "network-port": {
                    "ip-address": str(self.localIPAddr),
                    "ip-subnet-mask": str(self.localIPAddr.netmask),
                    "bacnet-ip-udp-port": self.localIPAddr.addrPort,
                    "network-number": None,
                    "fd-bbmd-address": sequence_to_json(HostNPort(self.bbmdAddress)),
                    "fd-subscription-lifetime": self.bbmdTTL,
                    "bacnet-ip-mode": mode,
                },
            }
            if mode == "bbmd":
                # bdt_json_seq = [f"BDTEntry({addr})" for addr in self.bdtable]
                cfg["network-port"]["bbmdBroadcastDistributionTable"] = self.bdtable

            _cfg = config(cfg)

            self.this_application = BAC0Application(
                _cfg, self.localIPAddr, json_file=self.json_file
            )
            if mode == "bbmd":
                self._log.info(f"Populating BDT with {self.bdtable}")
                self.this_application.populate_bdt()

            if mode == "foreign":
                self._log.info(
                    f"Registering as a foreign device to host {self.bbmdAddress} for {self.bbmdTTL} seconds"
                )
                self.this_application.register_as_foreign_device_to(
                    host=self.bbmdAddress, lifetime=self.bbmdTTL
                )

            self.log("Starting", level="debug")
            self._initialized = True

            try:
                Base._used_ips.add(self.localIPAddr)
                self.log(f"Registered as {app_type} | mode {mode}", level="info")
                self._started = True
            except OSError as error:
                self.log(f"Error opening socket: {error}", level="warning")
                raise InitializationError(f"Error opening socket: {error}")
            self.log("Running", level="debug")
        except OSError as error:
            self.log(f"an error has occurred: {error}", level="error")
            raise InitializationError(f"Error starting app: {error}")
            self.log("finally", level="debug")

    def register_foreign_device(self, addr=None, ttl=0):
        # self.this_application.register_to_bbmd(addr, ttl)
        raise NotImplementedError()

    def unregister_foreign_device(self):
        self.this_application.unregister_from_bbmd()

    def disconnect(self) -> asyncio.Task:
        task = asyncio.create_task(self._disconnect())
        return task

    async def _disconnect(self):
        """
        Stop the BACnet stack.  Free the IP socket.
        """
        self.log("Stopping All running tasks", level="debug")
        await stopAllTasks()
        self.log("Stopping BACnet stack", level="debug")
        # Freeing socket
        self.this_application.app.close()

        self._stopped = True  # Stop stack thread
        # self.t.join()
        self._started = False
        Base._used_ips.discard(self.localIPAddr)
        self.log("BACnet stopped", level="info")

    @property
    def routing_table(self):
        """
        Routing Table will give all the details about routers and how they
        connect BACnet networks together.

        It's a decoded presentation of what bacpypes.router_info_cache contains.

        Returns a dict with the address of routers as key.
        """

        class Router:
            def __init__(self, snet, address, dnets, path=None):
                self.source_network: int = snet
                self.address: Address = address
                self.destination_networks: set = dnets
                self.path: list = path

            def __repr__(self):
                return "Source Network: {} | Address: {} | Destination Networks: {} | Path: {}".format(
                    self.source_network,
                    self.address,
                    self.destination_networks,
                    self.path,
                )

        self._routers = {}

        self._ric = self.this_application.app.nsap.router_info_cache

        for router, dnets in self._ric.router_dnets.items():
            snet, address = router
            self._routers[str(address)] = Router(snet, address, dnets, path=[])
        for path, router_info in self._ric.path_info.items():
            router_address, router_status = router_info
            snet, dnet = path
            self._routers[str(router_address)].path.append((path, router_status))

        return self._routers

    @classmethod
    def extract_value_from_primitive_data(value):
        if isinstance(value, float):
            return float(value)
        # elif isinstance(value, Boolean):
        #    if value == int(1):
        #        return True
        #    else:
        #        return False
        elif isinstance(value, int):
            return int(value)
        elif isinstance(value, str):
            return str(value)
        else:
            return value



from bacpypes3.apdu import WritePropertyRequest
from bacpypes3.app import Application
from bacpypes3.basetypes import DeviceObjectPropertyReference, EventParameter
from bacpypes3.constructeddata import Any
from bacpypes3.pdu import Address

from BAC0.core.app.asyncApp import BAC0Application


# this class follows the same design as schedule/calendar
class EventEnrollment:
    def __make_event_parameters_request(
        self, destination, object_instance, event_parameters
    ):
        request = WritePropertyRequest(
            objectIdentifier=("eventEnrollment", object_instance),
            propertyIdentifier="eventParameters",
        )

        address = Address(destination)
        request.pduDestination = address
        request.propertyValue = Any(None)
        request.propertyValue.cast_in(event_parameters)
        request.priority = 15
        return request

    def __send_event_parameters_request(self, request, timeout=10):
        _this_application: BAC0Application = self.this_application
        _app: Application = _this_application.app

        self.log(f"{'- request:':>12} {request}", level="debug")

        _app.request(request)

        self._log.info(
            f"Event Parameters Write request sent to device : {request.pduDestination}"
        )

    # external interface
    def write_event_parameters(
        self, destination, event_enrollment_instance, event_parameters
    ):
        if type(event_parameters) is not EventParameter:
            self.log.warning("Data is not of type EventParameter.")
            return

        request = self.__make_event_parameters_request(
            destination=destination,
            object_instance=event_enrollment_instance,
            event_parameters=event_parameters,
        )
        self.__send_event_parameters_request(request)

    def __make_obj_prop_ref_request(self, destination, object_instance, obj_prop_ref):
        request = WritePropertyRequest(
            objectIdentifier=("eventEnrollment", object_instance),
            propertyIdentifier="objectPropertyReference",
        )

        address = Address(destination)
        request.pduDestination = address
        request.propertyValue = Any(None)
        request.propertyValue.cast_in(obj_prop_ref)
        request.priority = 15
        return request

    def __send_obj_prop_ref_request(self, request, timeout=10):
        _this_application: BAC0Application = self.this_application
        _app: Application = _this_application.app

        self.log(f"{'- request:':>12} {request}", level="debug")

        _app.request(request)

        self._log.info(
            "Object Property Reference Write request sent to device : {request.pduDestination}"
        )

    # external interface
    def write_obj_prop_ref(self, destination, event_enrollment_instance, obj_prop_ref):
        if type(obj_prop_ref) is not DeviceObjectPropertyReference:
            self.log.warning("Data is not of type ObjectPropertyReference.")
            return

        request = self.__make_obj_prop_ref_request(
            destination=destination,
            object_instance=event_enrollment_instance,
            obj_prop_ref=obj_prop_ref,
        )
        self.__send_obj_prop_ref_request(request)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
import typing as t

# --- standard Python modules ---
from datetime import time as dt_time

from bacpypes3.apdu import WritePropertyRequest
from bacpypes3.app import Application
from bacpypes3.basetypes import DailySchedule, TimeValue
from bacpypes3.constructeddata import Any, ArrayOf

# --- 3rd party modules ---
from bacpypes3.pdu import Address
from bacpypes3.primitivedata import Enumerated, Integer, Real

from ...core.app.asyncApp import BAC0Application
from ...core.utils.notes import note_and_log


@note_and_log
class Schedule:
    """
    Everything you need to write to a schedule
    """

    WeeklySchedule = ArrayOf(DailySchedule)
    schedules = {}
    days = [
        "monday",
        "tuesday",
        "wednesday",
        "thursday",
        "friday",
        "saturday",
        "sunday",
    ]

    schedule_example_multistate = {
        "states": {"Occupied": 1, "UnOccupied": 2, "Standby": 3, "Not Set": 4},
        "week": {
            "monday": [("1:00", "Occupied"), ("17:00", "UnOccupied")],
            "tuesday": [("2:00", "Occupied"), ("17:00", "UnOccupied")],
            "wednesday": [("3:00", "Occupied"), ("17:00", "UnOccupied")],
            "thursday": [("4:00", "Occupied"), ("17:00", "UnOccupied")],
            "friday": [("5:00", "Occupied"), ("17:00", "UnOccupied")],
            "saturday": [("6:00", "Occupied"), ("17:00", "UnOccupied")],
            "sunday": [("7:00", "Occupied"), ("17:00", "UnOccupied")],
        },
    }
    schedule_example_binary = {
        "states": {"inactive": 0, "active": 1},
        "week": {
            "monday": [("1:00", "active"), ("16:00", "inactive")],
            "tuesday": [("2:00", "active"), ("16:00", "inactive")],
            "wednesday": [("3:00", "active"), ("16:00", "inactive")],
            "thursday": [("4:00", "active"), ("16:00", "inactive")],
            "friday": [("5:00", "active"), ("16:00", "inactive")],
            "saturday": [("6:00", "active"), ("16:00", "inactive")],
            "sunday": [("7:00", "active"), ("16:00", "inactive")],
        },
    }
    schedule_example_analog = {
        "states": "analog",
        "week": {
            "monday": [("1:00", 22), ("18:00", 19)],
            "tuesday": [("2:00", 22), ("18:00", 19)],
            "wednesday": [("3:00", 22), ("18:00", 19)],
            "thursday": [("4:00", 22), ("18:00", 19)],
            "friday": [("5:00", 22), ("18:00", 19)],
            "saturday": [("6:00", 22), ("18:00", 19)],
            "sunday": [("7:00", 22), ("18:00", 19)],
        },
    }

    def create_weeklySchedule(self, dict_schedule, object_reference=None):
        """
        From a structured dict (see schedule_example), create a WeeklySchedule
        an ArrayOf(DailySchedule)
        """
        ds = dict_schedule
        if object_reference is not None:
            self.schedules[object_reference] = ds

        def _set_value(v):
            try:
                if dict_schedule["states"].lower() == "analog":
                    return Real(v)
            except AttributeError:
                if dict_schedule["states"] == {"inactive": 0, "active": 1}:
                    return Integer(dict_schedule["states"][v])
                else:
                    return Enumerated(dict_schedule["states"][v])

        daily_schedules = []
        for day in Schedule.days:
            list_of_events = dict_schedule["week"][day]
            _daily_schedule = [
                TimeValue(time=event[0], value=_set_value(event[1]))
                for event in list_of_events
            ]
            daily_schedules.append(DailySchedule(daySchedule=_daily_schedule))
        return Schedule.WeeklySchedule(daily_schedules)

    def make_weeklySchedule_request(self, destination, object_instance, weeklySchedule):
        request = WritePropertyRequest(
            objectIdentifier=("schedule", object_instance),
            propertyIdentifier="weeklySchedule",
        )

        address = Address(destination)
        request.pduDestination = address
        request.propertyValue = Any(weeklySchedule)
        request.priority = 15
        return request

    def send_weeklyschedule_request(self, request, timeout=10):
        _this_application: BAC0Application = self.this_application
        _app: Application = _this_application.app

        self.log(f"{'- request:':>12} {request}", level="debug")

        _app.request(request)

        self._log.info(
            f"Schedule Write request sent to device : {request.pduDestination}"
        )

    def write_weeklySchedule(self, destination, schedule_instance, schedule):
        weeklyschedule = self.create_weeklySchedule(schedule)
        request = self.make_weeklySchedule_request(
            destination=destination,
            object_instance=schedule_instance,
            weeklySchedule=weeklyschedule,
        )
        self.send_weeklyschedule_request(request)

    async def read_weeklySchedule(self, address, schedule_instance):
        """
        This function will turn the weeklySchedule received into a
        human readable dict.
        This dict can then be modified and used to write back to controller
        using bacnet.write_weeklySchedule
        """
        try:
            (
                _schedule,
                object_references,
                reliability,
                priority,
                presentValue,
            ) = await self.readMultiple(
                "{} schedule {} weeklySchedule listOfObjectPropertyReferences reliability priorityForWriting presentValue".format(
                    address, schedule_instance
                )
            )

        except Exception:
            # Read Multiple not supported... try single prop read
            def _read(prop):
                return self.read(f"{address} schedule {schedule_instance} {prop}")

            try:
                _schedule = _read("weeklySchedule")
                object_references = _read("listOfObjectPropertyReferences")
                reliability = _read("reliability")
                priority = _read("priorityForWriting")
                presentValue = _read("presentValue")

            except Exception:
                raise ()

        schedule: t.Dict[str, t.Union[t.Dict, t.List]] = {}
        _state_text: t.Union[str, range, t.List[str], None] = None
        offset_MV = 0 if len(object_references) == 0 else 1

        try:
            first_obj_id = object_references[0]
            obj_type, obj_instance = first_obj_id.objectIdentifier
            if "binary" in str(obj_type):
                _state_text = ["inactive", "active"]
            elif "multi" in str(obj_type):
                _state_text = await self.read(
                    f"{address} {obj_type} {obj_instance} stateText"
                )
            elif "analog" in str(obj_type):
                _state_text = "analog"

            schedule["object_references"] = [
                objectId.objectIdentifier for objectId in object_references
            ]
            schedule["references_names"] = [
                await self.read(
                    "{} {} {} objectName".format(
                        address, each.objectIdentifier[0], each.objectIdentifier[1]
                    )
                )
                for each in object_references
            ]
        except Exception as error:
            self.log(f"Error {error}", level="error")
            # Not used in device, not linked...
            _state_text = range(255)
            schedule["object_references"] = []
            schedule["references_names"] = []

        # re-ordered to make type inference possible, but not sure the logic here
        # is 100% sound (ignored type warnings both look like actual problems).
        # keeping logic intact and ignoring warnings for now
        sched_states = {}
        if _state_text == ["inactive", "active"]:
            for i, each in enumerate(_state_text):
                sched_states[each] = i
            presentValue = "{} ({})".format(
                list(sched_states.keys())[int(presentValue.get_value())],
                presentValue.get_value(),
            )
        elif _state_text == "analog":
            sched_states = _state_text  # type: ignore[assignment]
            if presentValue is not None:
                presentValue = f"{presentValue.get_value()}"
        else:
            try:
                for i, each in enumerate(_state_text):  # type: ignore[arg-type]
                    sched_states[each] = i + offset_MV
                presentValue = "{} ({})".format(
                    list(sched_states.keys())[
                        int(presentValue.get_value()) - offset_MV
                    ],
                    presentValue.get_value(),
                )
            except TypeError:
                presentValue = presentValue.get_value()

        schedule["states"] = sched_states
        schedule["reliability"] = reliability
        schedule["priority"] = priority
        schedule["presentValue"] = presentValue
        schedule["week"] = self.decode_weeklySchedule(_schedule, _state_text, offset_MV)

        return schedule

    def decode_weeklySchedule(self, weeklySchedule, states, offset_MV):
        week = {}
        for i, day in enumerate(Schedule.days):
            week[day] = self.decode_dailySchedule(weeklySchedule[i], states, offset_MV)
        return week

    def decode_dailySchedule(self, dailySchedule, states, offset_MV):
        events = []
        for each in dailySchedule.daySchedule:
            hour, minute, second, hundreth = each.time
            _time = dt_time(hour, minute, second, hundreth).strftime("%H:%M")
            try:
                _value = None
                if each.value.get_value() is not None:
                    if states == ["inactive", "active"]:
                        _value = states[int(each.value.get_value())]
                    elif states == "analog":
                        _value = float(each.value.get_value())
                    else:
                        _value = states[int(each.value.get_value()) - offset_MV]
                events.append((_time, _value))
            except IndexError:
                events.append((_time, f"??? ({each.value.get_value()})"))
        return events



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
Points.py - Definition of points so operations on Read results are more convenient.
"""

import asyncio
import hashlib
from collections import namedtuple


# --- standard Python modules ---
from datetime import datetime

from bacpypes3.basetypes import EngineeringUnits

from ...tasks.Match import Match_Value

# --- this application's modules ---
from ..utils.notes import note_and_log
from ..utils.lookfordependency import pandas_if_available

_PANDAS, pd, _, _ = pandas_if_available()
# ------------------------------------------------------------------------------


class VirtualDeviceProperties(object):
    """
    This serves as a container for device properties
    """

    def __init__(self):
        self.name = "Virtual Device"
        self.address = "local"
        self.device_id = None
        self.network = None
        self.pollDelay = None
        self.objects_list = None
        self.pss = None
        self.multistates = None
        self.db_name = None
        self.segmentation_supported = True
        self.history_size = None
        self.bacnet_properties = None

    def __repr__(self):
        return f"{self.asdict}"

    @property
    def asdict(self):
        return self.__dict__


class VirtualDevice(object):
    def __init__(self):
        self.properties = VirtualDeviceProperties()


class VirtualPointProperties(object):
    """
    A container for VirtualPoint properties
    """

    def __init__(self):
        self.device = None
        self.name = None
        self.type = "analog-virtual"
        self.address = None
        self.description = ""
        self.units_state = ""
        self.simulated = False
        self.overridden = False
        self.priority_array = None
        self.history_size = None
        self.bacnet_properties = {}
        self.status_flags = None
        self._df = None

    def __repr__(self):
        return f"{self.asdict}"

    @property
    def asdict(self):
        return self.__dict__


@note_and_log
class VirtualPoint:
    """
    Virtual points could be used to show calculation result on Bokeh
    A function is passed at creation time and this function must return a pandas Serie

    I chose to create those points without using BACnet classes because Virtual
    points are meant to be used on BAC0.device and creating fake BACnet objects
    could lead to potential confusion.
    """

    def __init__(
        self,
        name,
        device=None,
        object_type="analogVirtual",
        initial_value=None,
        history_fn=None,
        description=None,
        units="No Units",
        tags=[],
    ):
        self.properties = VirtualPointProperties()
        self.properties.type = object_type
        self.properties.device = VirtualDevice() if device is None else device
        if description is None:
            raise ValueError("Please provide description")
        if not _PANDAS:
            raise ImportError("Pandas required to use VirtualPoints")
        self.name = name
        self.properties.name = name
        self.properties.address = int(hashlib.sha256(name.encode()).hexdigest(), 16) % (
            10**8
        )
        # 8 digits address based on the name. this way I don't have to keep track of numbers and if the same name is used, we should get something similar and the InfluxDB trend will mark the point as the same.
        self.properties.description = description
        self.properties.units_state = (
            EngineeringUnits(units) if object_type == "analogVirtual" else units
        )
        self.tags = tags
        self._history_fn = history_fn

        self._history = namedtuple("_history", ["timestamp", "value"])
        self._history.timestamp = []
        self._history.value = []

        self._match_task = namedtuple("_match_task", ["task", "running"])
        self._match_task.task = None
        self._match_task.running = False

        self.fake_pv = None
        if initial_value:
            self.fake_pv = float(initial_value)
            self._trend(float(initial_value))

    def chart(self, remove=False):
        """
        Add point to the bacnet trending list
        """
        self.log("Use bacnet.add_trend(point) instead", level="warning")

    async def _set(self, value):
        if value == "auto":
            pass
        elif self._history_fn is None:
            self.fake_pv = float(value)
            self._trend(float(value))
        else:
            self._log.warning(
                "Point is configured as a function of other points. You can't set a new value"
            )

    def _trend(self, res):
        self._history.timestamp.append(datetime.now().astimezone())
        self._history.value.append(res)
        if self.properties.device.properties.network.database:
            self.properties.device.properties.network.database.prepare_point([self])

        if self.properties.history_size is None:
            return
        else:
            if self.properties.history_size < 1:
                self.properties.history_size = 1
            if len(self._history.timestamp) >= self.properties.history_size:
                try:
                    self._history.timestamp = self._history.timestamp[
                        -self.properties.history_size :
                    ]
                    self._history.value = self._history.value[
                        -self.properties.history_size :
                    ]
                    assert len(self._history.timestamp) == len(self._history.value)
                except Exception:
                    self._log.exception("Can't append to history")

    @property
    def lastTimestamp(self):
        """
        returns: last timestamp read
        """
        if _PANDAS:
            last_val = self.history.dropna()
            last_val_clean = None if len(last_val) == 0 else last_val.index[-1]
            return last_val_clean
        else:
            return self._history.timestamp[-1]

    @property
    async def value(self):
        """
        Retrieve value of the point
        """
        return self.lastValue

    @property
    def lastValue(self):
        """
        returns: last value read
        """
        if _PANDAS:
            return self.history.dropna().iloc[-1]
        else:
            return self._history.value[-1]

    @property
    def history(self):
        """
        returns : (pd.Series) containing timestamp and value of all readings
        """
        if self._history_fn is not None:
            return self._history_fn()
        else:
            if not _PANDAS:
                return dict(zip(self._history.timestamp, self._history.value))
            idx = self._history.timestamp.copy()
            his_table = pd.Series(index=idx, data=self._history.value[: len(idx)])
            del idx
            his_table.name = ("{}/{}").format(
                self.properties.device.properties.name, self.properties.name
            )
            his_table.units = self.properties.units_state
            his_table.states = "virtual"

            his_table.description = self.properties.description

            his_table.datatype = self.properties.type
            return his_table

    def match_value(self, value, *, delay=5, use_last_value=False):
        asyncio.create_task(
            self._match_value(value=value, delay=delay, use_last_value=use_last_value)
        )

    async def _match_value(self, value, *, delay=5, use_last_value=False):
        """
        This allow functions like :
            device['point'].match('value')

        A sensor will follow a calculation...
        """
        if self._match_task.task is None or not self._match_task.running:
            self._match_task.task = Match_Value(
                value=value, point=self, delay=delay, use_last_value=use_last_value
            )
            self._match_task.task.start()
            self._match_task.running = True

        elif self._match_task.running and delay > 0:
            await self._match_task.task.stop()
            self._match_task.running = False
            await asyncio.sleep(1)

            self._match_task.task = Match_Value(
                value=value, point=self, delay=delay, use_last_value=use_last_value
            )
            self._match_task.task.start()
            self._match_task.running = True

        elif self._match_task.running and delay == 0:
            await self._match_task.task.stop()
            self._match_task.running = False

        else:
            raise RuntimeError("Stop task before redefining it")

    def tag(self, tag_id, tag_value, lst=None):
        """
        Add tag to point. Those tags can be used to make queries,
        add information, etc.
        They will be included if InfluxDB is used.
        """
        if lst is None:
            self.tag.append((tag_id, tag_value))
        else:
            for each in lst:
                tag_id, tag_value = each
                self.tag.append((tag_id, tag_value))

    def __repr__(self):
        return "{}/{} : {:.2f} {}".format(
            self.properties.device.properties.name,
            self.properties.name,
            float(self.lastValue),
            self.properties.units_state,
        )

    @property
    def asdict(self):
        return self.__dict__

    def __add__(self, other):
        return self.lastValue + other

    __radd__ = __add__

    def __sub__(self, other):
        return self.lastValue - other

    def __rsub__(self, other):
        return other - self.lastValue

    def __mul__(self, other):
        return self.lastValue * other

    __rmul__ = __mul__

    def __truediv__(self, other):
        return self.lastValue / other

    def __rtruediv__(self, other):
        return other / self.lastValue

    def __lt__(self, other):
        return self.lastValue < other

    def __le__(self, other):
        return self.lastValue <= other

    def __eq__(self, other):
        return self.lastValue == other

    def __gt__(self, other):
        return self.lastValue > other

    def __ge__(self, other):
        return self.lastValue >= other



import pytest

from BAC0.core.io.Write import write_pattern


@pytest.mark.parametrize(
    "req",
    [
        "303:5 binaryOutput:2132 presentValue inactive - 8",
        "192.168.1.149 binaryOutput:2132 presentValue inactive - 8",
        "192.168.1.149 binaryOutput:2132 presentValue 0.1 - 8",
        "192.168.1.149 binaryOutput 2132 presentValue 0.1234 - 8",
        "303:5 analogInput 1095 presentValue 15.3 - 8"
        "303:5 binaryOutput 2132 presentValue inactive - 8",
        "303:5 @obj_142:2132 presentValue inactive - 8",
        "303:5 @obj_142:2132 @prop_345 inactive - 8",
        "303:5 @obj_142:2132 @prop_345 inactive",
        "303:5 binaryInput:1095 presentValue active",
        "303:5 analog-input 1095 presentValue 10.3",
        "303:5 analogInput:1095 presentValue 15.3",
        # "303:5 analogInput:1095",
        # "303:5 analog-input:1095",
        "303:5 analog-input 1095",
    ],
)
@pytest.mark.asyncio
async def test_pattern(network_and_devices, req):
    # async for resources in network_and_devices:
    # loop, bacnet, device_app, device30_app, test_device, test_device_30 = resources
    match = write_pattern.search(req)
    assert match is not None, f"Pattern did not match for request: {req}"
    address = match.group("address")
    objId = match.group("objId")
    prop_id = match.group("propId")
    value = match.group("value")
    indx = match.group("indx")
    priority = match.group("priority")
    assert address is not None, "Address not found"
    assert objId is not None, "Object ID not found"
    assert prop_id is not None, "Property ID not found"
    assert value is not None, "Value not found"



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
Reinitialize.py - creation of ReinitializeDeviceRequest

"""
from bacpypes3.apdu import (
    ReinitializeDeviceRequest,
    ReinitializeDeviceRequestReinitializedStateOfDevice,
)


# --- 3rd party modules ---
from bacpypes3.pdu import Address
from bacpypes3.app import Application
from bacpypes3.primitivedata import CharacterString

from ..io.IOExceptions import ApplicationNotStarted
from ..utils.notes import note_and_log
from ...core.app.asyncApp import BAC0Application

# --- standard Python modules ---


@note_and_log
class Reinitialize:
    """
    Mixin to support Reinitialize from BAC0 to other devices
    """

    def reinitialize(self, address=None, password=None, state="coldstart"):
        """
        Will send reinitialize request
        """
        _this_application: BAC0Application = self.this_application
        _app: Application = _this_application.app
        if not self._started:
            raise ApplicationNotStarted("BACnet stack not running - use startApp()")

        if not address:
            raise ValueError("Provide address for request")

        # build a request
        request = ReinitializeDeviceRequest()
        request.reinitializedStateOfDevice = getattr(
            ReinitializeDeviceRequestReinitializedStateOfDevice, state
        )
        request.pduDestination = Address(address)
        request.password = CharacterString(password)

        self.log(f"{'- request:':>12} {request}", level="debug")
        _app.request(request)

        self.log(f"Reinitialize request sent to device : {address}", level="info")



def tec_short_point_list(unit_type="2-pipe"):
    """
    unit_type can be :
        - 4-pipe
        - 2-pipe
        - VAV
    """
    _lst = [
        ("binaryInput", 30827),
        ("binaryInput", 30828),
        ("binaryOutput", 86908),
        ("binaryOutput", 86909),
        ("binaryOutput", 86910),
        ("binaryOutput", 86911),
        ("binaryOutput", 86912),
        ("binaryOutput", 87101),
        ("binaryOutput", 87102),
        ("multiStateValue", 29501),
        ("multiStateValue", 29500),
        ("multiStateValue", 29509),
        ("multiStateValue", 29517),
        ("multiStateValue", 29518),
        ("multiStateValue", 29519),
        ("multiStateValue", 29520),
        ("multiStateValue", 29524),
        ("multiStateValue", 29525),
        ("multiStateValue", 29527),
        ("multiStateValue", 29712),
        ("multiStateValue", 29700),
        ("multiStateValue", 29709),
        ("multiStateValue", 29708),
        ("analogValue", 29505),
        # ("analogValue", 29502),
        # ("analogValue", 29503),
        ("analogValue", 29504),
        ("analogValue", 29506),
        ("analogValue", 29507),
        ("analogValue", 29508),
        ("analogValue", 29515),
        ("analogValue", 29522),
        ("analogValue", 29529),
        ("analogValue", 29530),
        ("analogValue", 29532),
        ("analogValue", 29701),
        ("analogValue", 29703),
        ("analogValue", 29705),
        ("analogValue", 29706),
        ("analogValue", 29707),
        ("analogValue", 29714),
        ("analogValue", 29717),
        ("analogValue", 29725),
        ("analogValue", 29726),
        ("analogValue", 29727),
        ("analogOutput", 86905),
        ("multiStateValue", 6),
        ("trendLog", 101010),
    ]
    if unit_type == "4-pipe":
        _lst.append(("analogOutput", 86914))
        _lst.append(("analogOutput", 86915))

    return _lst



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
Match.py - verify a point's status matches its commanded value.

Example:
    Is a fan commanded to 'On' actually 'running'?
"""

# --- standard Python modules ---
# --- 3rd party modules ---

import asyncio

from ..core.utils.notes import note_and_log
from ..core.io.IOExceptions import NotReadyError

# --- this application's modules ---
from .TaskManager import Task

# ------------------------------------------------------------------------------


@note_and_log
class Match(Task):
    """
    Match two properties of a BACnet Object (i.e. a point status with its command).
    """

    def __init__(self, status=None, command=None, delay=5, name=None):
        self._log.debug(
            f"Creating Match task for {command} and {status}. Delay : {delay}"
        )
        if not name:
            name = "Match on " + status.properties.name
        self.command = command
        self.status = status
        Task.__init__(self, delay=delay, name=name)

    async def task(self):
        if (
            self.status.properties.network.initialized is False
            or self.status.properties.network.initialized is None
            or self.status is None
        ):
            raise NotReadyError(f"{self.status} is not ready")
        if (
            self.command.properties.network.initialized is False
            or self.command.properties.network.initialized is None
            or self.command is None
        ):
            raise NotReadyError(f"{self.command} is not ready")
        try:
            if self.status.history[-1] != self.command.history[-1]:
                _val = (
                    self.command.history[-1].split(":")[1]
                    if ":" in self.command.history[-1]
                    else self.command.history[-1]
                )
                self.log(f"Match value is {_val}", level="debug")
                await self.status._setitem(_val.replace(" ", ""))
        except (NotReadyError, TypeError) as error:
            self.log(
                f"Problem executing match value task {self.status.name} -> {self.command.name} : {error}",
                level="warning",
            )
        except Exception as error:
            self._log.error(
                f"Something wrong matching {self.command.properties.name} and {self.status.properties.name}... try again next time...\nError:{error}"
            )
            await asyncio.sleep(1)

    async def stop(self):
        await self.status._setitem("auto")
        super().stop()


@note_and_log
class Match_Value(Task):
    """
    Verify a point's Present_Value equals the given value after a delay of X seconds.
    Thus giving the BACnet controller (and connected equipment) time to respond to the
    command.

        Match_Value(On, <AI:1>, 5)

    i.e. Does Fan value = On after 5 seconds.
    """

    def __init__(
        self, value=None, point=None, delay=5, name=None, use_last_value=False
    ):
        self.log(f"Creating MatchValue task for {value} and {point}", level="debug")
        # if not isinstance(value, (float, int, str, bool)) or not hasattr(self.value, "__call__"):
        #    raise ValueError("Value must be a float, int, str or bool OR must be a callable function that returns one of these types.")
        self.value = value
        self.point = point
        self.use_last_value = use_last_value
        if not name:
            name = "Match_Value on " + point.properties.name
        Task.__init__(self, delay=delay, name=name)

    async def task(self):
        if (
            self.point.properties.device.initialized is False
            or self.point.properties.device.initialized is None
            or self.point is None
        ):
            raise NotReadyError(f"{self.point} is not ready")
        try:
            if self.use_last_value:
                _point = self.point.lastValue
            else:
                _point = await self.point.value
            value = self.value() if hasattr(self.value, "__call__") else self.value
            if value != _point:
                await self.point._set(value=value)
        except (NotReadyError, TypeError) as error:
            self.log(
                f"Problem executing match value task on {self.point.properties.name} -> {value}: {error}",
                level="warning",
            )
            await asyncio.sleep(1)
        except Exception as error:
            self.log(
                f"Something is wrong matching {self.value} and {self.point.properties.name}... try again next time {error}",
                level="error",
            )
            await asyncio.sleep(1)

    async def _before_stop(self):
        try:
            await self.point._set("auto")
        except (ValueError, TypeError):
            self.log(
                "Could not set {} to auto. If this is a network input, it is normal as we didn't have to override or simulate".format(
                    self.point
                ),
                level="warning",
            )

    async def stop(self):
        await self._before_stop()
        super().stop()



# -*- coding: utf-8 -*-
"""
Notes and logger decorator to be used on class
This will add a "notes" object to the class and will allow
logging feature at the same time.
Goal is to be able to access quickly to important informations for
the web interface.
"""
import inspect
import logging
import os
import sys
import typing as t

# --- standard Python modules ---
from collections import namedtuple
from datetime import datetime
from logging import FileHandler, Logger
from os.path import expanduser, join

# --- 3rd party modules ---
from ...core.utils.lookfordependency import rich_if_available, pandas_if_available

RICH, rich = rich_if_available()
if RICH:
    from rich.logging import RichHandler

_PANDAS, pd, _, _ = pandas_if_available()


class LogList:
    LOGGERS: t.List[Logger] = []


def convert_level(level):
    if not level:
        return None
    _valid_levels = [
        logging.DEBUG,
        logging.INFO,
        logging.WARNING,
        logging.ERROR,
        logging.CRITICAL,
    ]
    if level in _valid_levels:
        return level
    if level.lower() == "info":
        return logging.INFO
    elif level.lower() == "debug":
        return logging.DEBUG
    elif level.lower() == "warning":
        return logging.WARNING
    elif level.lower() == "error":
        return logging.ERROR
    elif level.lower() == "critical":
        return logging.CRITICAL
    raise ValueError(f"Wrong log level use one of the following : {_valid_levels}")


def update_log_level(
    level=None, *, log_file=None, stderr=None, stdout=None, log_this=True
):
    """
    Typical usage ::
        # Silence (use CRITICAL so not much messages will be sent)
        BAC0.log_level('silence')
        # Verbose
        BAC0.log_level('info')
        # Default, Info on console....but Warning in file
        BAC0.log_level(file='warning', stdout='info', stderr='critical')
        # Debug in file and console... this is a bad idea as the console will be filled
        BAC0.log_level(file='debug', stdout='debug', stderr='critical')

        # Preferably, debug in the file, keep console limited to info
        BAC0.log_level('debug')
        # OR
        BAC0.log_level(file='debug', stdout='info', stderr='critical')


    Giving only one parameter will set file and console to the same level.
    I tend to keep stderr CRITICAL

    """
    update_log_file_lvl = False
    update_stderr_lvl = False
    update_stdout_lvl = False

    if level:
        logging.getLogger("BAC0_Root.BAC0.scripts.Base.Base").disabled = False
        if level.lower() == "silence":
            log_file_lvl = logging.CRITICAL
            stderr_lvl = logging.CRITICAL
            stdout_lvl = logging.CRITICAL
            update_log_file_lvl = True
            update_stderr_lvl = True
            update_stdout_lvl = True
            logging.getLogger("BAC0_Root.BAC0.scripts.Base.Base").disabled = True
        elif level.lower() == "default":
            log_file_lvl = logging.WARNING
            stderr_lvl = logging.CRITICAL
            stdout_lvl = logging.INFO
            update_log_file_lvl = True
            update_stderr_lvl = True
            update_stdout_lvl = True
        elif level.lower() == "debug":
            log_file_lvl = logging.DEBUG
            stdout_lvl = logging.INFO
            update_log_file_lvl = True
            update_stdout_lvl = True
        else:
            level = convert_level(level)
            log_file_lvl = level
            stdout_lvl = level
            update_log_file_lvl = True
            update_stdout_lvl = True

    else:
        if log_file:
            log_file_lvl = convert_level(log_file)
            update_log_file_lvl = True
        if stderr:
            stderr_lvl = convert_level(stderr)
            update_stderr_lvl = True
        if stdout:
            stdout_lvl = convert_level(stdout)
            update_stdout_lvl = True

    # Choose Base as logger for this task
    if log_this:
        BAC0_logger = logging.getLogger("BAC0_Root.BAC0.scripts.Base.Base")

    for each in LogList.LOGGERS:
        for handler in each.handlers:
            if update_log_file_lvl and handler.get_name() == "file_handler":
                handler.setLevel(log_file_lvl)
                if log_this:
                    BAC0_logger.warning(
                        "Changed log level of file to {}".format(
                            logging.getLevelName(log_file_lvl)
                        )
                    )
            elif update_stdout_lvl and handler.get_name() == "stdout":
                handler.setLevel(stdout_lvl)
                if log_this:
                    BAC0_logger.warning(
                        "Changed log level of console stdout to {}".format(
                            logging.getLevelName(stdout_lvl)
                        )
                    )
            elif update_stderr_lvl and handler.get_name() == "stderr":
                handler.setLevel(stderr_lvl)
                if log_this:
                    BAC0_logger.warning(
                        "Changed log level of console stderr to {}".format(
                            logging.getLevelName(stderr_lvl)
                        )
                    )


def note_and_log(cls):
    """
    This will be used as a decorator on class to activate
    logging and store messages in the variable cls._notes
    This will allow quick access to events in the web app.
    A note can be added to cls._notes without logging if passing
    the argument log=false to function note()
    Something can be logged without addind a note using function log()
    """
    if hasattr(cls, "DEBUG_LEVEL"):
        if cls.DEBUG_LEVEL == "debug":
            file_level = logging.DEBUG
            console_level = logging.DEBUG
        elif cls.DEBUG_LEVEL == "info":
            file_level = logging.INFO
            console_level = logging.INFO
    else:
        file_level = logging.WARNING
        console_level = logging.INFO
    # Notes object
    cls._notes = namedtuple("_notes", ["timestamp", "notes"])
    cls._notes.timestamp = []
    cls._notes.notes = []

    # Defining log object
    cls.logname = f"{cls.__module__} | {cls.__name__}"
    cls._log = logging.getLogger(f"BAC0_Root.{cls.__module__}.{cls.__name__}")

    # Set level to debug so filter is done by handler
    cls._log.setLevel(logging.DEBUG)

    # Console Handler
    if RICH:
        ch = RichHandler()
        ch2 = RichHandler()
    else:
        ch = logging.StreamHandler(sys.stderr)
        ch2 = logging.StreamHandler(sys.stdout)

    ch.set_name("stderr")
    ch.setLevel(logging.CRITICAL)
    ch2.set_name("stdout")
    ch2.setLevel(console_level)

    formatter = logging.Formatter("{asctime} - {levelname:<8}| {message}", style="{")

    # File Handler
    _PERMISSION_TO_WRITE = True
    logUserPath = expanduser("~")
    logSaveFilePath = join(logUserPath, ".BAC0")

    logFile = join(logSaveFilePath, "BAC0.log")
    try:
        if not os.path.exists(logSaveFilePath):
            os.makedirs(logSaveFilePath)
        fh = FileHandler(logFile)
        fh.set_name("file_handler")
        fh.setLevel(file_level)
        fh.setFormatter(formatter)

    except OSError:
        _PERMISSION_TO_WRITE = False

    ch.setFormatter(formatter)
    ch2.setFormatter(formatter)
    # Add handlers the first time only...
    if not len(cls._log.handlers):
        if _PERMISSION_TO_WRITE:
            cls._log.addHandler(fh)
        cls._log.addHandler(ch)
        cls._log.addHandler(ch2)

    LogList.LOGGERS.append(cls._log)

    def log_title(self, title, args=None, width=35):
        cls._log.debug("")
        cls._log.debug("#" * width)
        cls._log.debug(f"# {title}")
        cls._log.debug("#" * width)
        if args:
            cls._log.debug(f"{args!r}")
            cls._log.debug("#" * 35)

    def log_subtitle(self, subtitle, args=None, width=35):
        cls._log.debug("")
        cls._log.debug("=" * width)
        cls._log.debug(f"{subtitle}")
        cls._log.debug("=" * width)
        if args:
            cls._log.debug(f"{args!r}")
            cls._log.debug("=" * width)

    def log(self, note, *, level: t.Union[str, int] = logging.DEBUG):
        """
        Add a log entry...no note
        """
        if not note:
            raise ValueError("Provide something to log")
        if isinstance(level, str):
            level = convert_level(level)
        if level == logging.INFO:
            note = f"{note}"
        else:
            caller_frame = inspect.stack()[1]
            module = inspect.getmodule(caller_frame[0])
            module_name = module.__name__ if module else "unknown"
            note = f"{cls.logname} | {module_name} | {note}"
        cls._log.log(level, note)

    def note(self, note, *, level=logging.INFO, log=True):
        """
        Add note to the object. By default, the note will also
        be logged
        :param note: (str) The note itself
        :param level: (logging.level)
        :param log: (boolean) Enable or disable logging of note
        """
        if not note:
            raise ValueError("Provide something to log")
        note = f"{cls.logname} | {note}"
        cls._notes.timestamp.append(datetime.now().astimezone())
        cls._notes.notes.append(note)
        if log:
            cls.log(level, note)

    @property
    def notes(self):
        """
        Retrieve notes list as a Pandas Series
        """
        if not _PANDAS:
            return dict(zip(self._notes.timestamp, self._notes.notes))
        return pd.Series(self._notes.notes, index=self._notes.timestamp)

    def clear_notes(self):
        """
        Clear notes object
        """
        cls._notes.timestamp = []
        cls._notes.notes = []

    # Add the functions to the decorated class
    cls.clear_notes = clear_notes
    cls.note = note
    cls.notes = notes
    cls.log = log
    cls.log_title = log_title
    cls.log_subtitle = log_subtitle
    return cls



#!/usr/BIn/env python
# -*- coding utf-8 -*-
from typing import AsyncGenerator
import pytest

"""
Test Bacnet communication with another device
"""

CHANGE_DELTA_AI = 99.90
CHANGE_DELTA_AO = 89.90
CHANGE_DELTA_AV = 79.90
TOLERANCE = 0.01
BINARY_TEST_STATE_STR1 = False
BINARY_TEST_STATE_STR2 = False
BINARY_TEST_STATE_BOOL = False
CHARACTERSTRINGVALUE = "test"


@pytest.mark.asyncio
async def test_ReadAnalog(network_and_devices: AsyncGenerator):
    async for resources in network_and_devices:
        loop, bacnet, device_app, device30_app, test_device, test_device_30 = resources

        await test_device["AV"].value
        assert (test_device["AV"].lastValue - CHANGE_DELTA_AV) < TOLERANCE

        # assert not test_device["MSV"] == 1

        assert test_device["BIG-ALARM"] == "Normal"
        await test_device["AI"].value
        assert (test_device["AI"].lastValue - CHANGE_DELTA_AI) < TOLERANCE
        await test_device["AO"].value
        assert (test_device["AO"].lastValue - CHANGE_DELTA_AO) < TOLERANCE

        # assert test_device["CS_VALUE"] == CHARACTERSTRINGVALUE


@pytest.mark.asyncio
async def test_ReadBinary(network_and_devices: AsyncGenerator):
    async for resources in network_and_devices:
        loop, bacnet, device_app, device30_app, test_device, test_device_30 = resources
        await test_device["BV-1"].value
        # assert test_device["BV-1"] is False
        print(test_device["BV-1"])
        # assert test_device["BV-1"] == BINARY_TEST_STATE
        # assert test_device["CS_VALUE"] == CHARACTERSTRINGVALUE
        assert test_device["BI"] == BINARY_TEST_STATE_STR1

        assert test_device["BO"] == BINARY_TEST_STATE_STR2
        assert test_device["BO-1"] == BINARY_TEST_STATE_BOOL
